from __future__ import division
from __future__ import unicode_literals

import pytz

_type = type

import datetime
from decimal import Decimal

from sqlalchemy import Column, VARCHAR, INTEGER, ForeignKey, BIGINT, Index
from sqlalchemy.orm import relation
from sqlalchemy.orm.exc import MultipleResultsFound

from ._commodity_helper import quandl_fx
from .._common import CallableList, GncConversionError
from .._common import GnucashException, hybrid_property_gncnumeric
from .._declbase import DeclarativeBaseGuid
from ..sa_extra import _DateAsDateTime
from ..yahoo_client import get_latest_quote, download_quote


class GncCommodityError(GnucashException):
    pass


class GncPriceError(GnucashException):
    pass


class Price(DeclarativeBaseGuid):
    """
    A single Price for a commodity.

    Attributes:
        commodity (:class:`Commodity`): commodity to which the Price relates
        currency (:class:`Commodity`): currency in which the Price is expressed
        date (:class:`datetime.date`): date object representing the day at which the price is relevant
        source (str): source of the price
        type (str): last, ask, bid, unknown, nav
        value (:class:`decimal.Decimal`): the price itself
    """
    __tablename__ = 'prices'

    __table_args__ = {}

    # column definitions
    commodity_guid = Column('commodity_guid', VARCHAR(length=32), ForeignKey('commodities.guid'), nullable=False)
    currency_guid = Column('currency_guid', VARCHAR(length=32), ForeignKey('commodities.guid'), nullable=False)
    date = Column('date', _DateAsDateTime(neutral_time=False), nullable=False)
    source = Column('source', VARCHAR(length=2048))
    type = Column('type', VARCHAR(length=2048))

    _value_num = Column('value_num', BIGINT(), nullable=False)
    _value_denom = Column('value_denom', BIGINT(), nullable=False)
    value = hybrid_property_gncnumeric(_value_num, _value_denom)

    # relation definitions
    commodity = relation('Commodity',
                         back_populates="prices",
                         foreign_keys=[commodity_guid],
                         )
    currency = relation('Commodity',
                        foreign_keys=[currency_guid],
                        )

    def __init__(self,
                 commodity,
                 currency,
                 date,
                 value,
                 type="unknown",
                 source="user:price"):
        self.commodity = commodity
        self.currency = currency
        assert _type(date) is datetime.date
        self.date = date
        self.value = value
        self.type = type
        self.source = source

    def __str__(self):
        return "Price<{:%Y-%m-%d} : {} {}/{}>".format(self.date,
                                                       self.value,
                                                       self.currency.mnemonic,
                                                       self.commodity.mnemonic)

    def object_to_validate(self, change):
        if change[-1] != "deleted":
            yield self

    def validate(self):
        # check uniqueness of namespace/mnemonic
        try:
            self.book.query(Price).filter_by(commodity=self.commodity,
                                             currency=self.currency,
                                             source=self.source,
                                             date=self.date).one()
        except MultipleResultsFound:
            raise ValueError("{} already exists in this book".format(self))


class Commodity(DeclarativeBaseGuid):
    """
    A GnuCash Commodity.

    Attributes:
        cusip (str): cusip code
        fraction (int): minimal unit of the commodity (e.g. 100 for 1/100)
        namespace (str): CURRENCY for currencies, otherwise any string to group multiple commodities together
        mnemonic (str): the ISO symbol for a currency or the stock symbol for stocks (used for online quotes)
        quote_flag (int): 1 if piecash/GnuCash quotes will retrieve online quotes for the commodity
        quote_source (str): the quote source for GnuCash (piecash always use yahoo for stock and quandl for currencies
        quote_tz (str): the timezone to assign on the online quotes
        base_currency (:class:`Commodity`): The base_currency for a commodity:

          - if the commodity is a currency, returns the "default currency" of the book (ie the one of the root_account)
          - if the commodity is not a currency, returns the currency encoded in the quoted_currency slot

        accounts (list of :class:`piecash.core.account.Account`): list of accounts which have the commodity as commodity
        transactions (list of :class:`piecash.core.transaction.Transaction`): list of transactions which have the commodity as currency
        prices (iterator of :class:`Price`): iterator on prices related to the commodity (it is a sqlalchemy query underneath)

    """
    __tablename__ = 'commodities'

    __table_args__ = (Index('_unique_cdty',
                            'namespace', 'mnemonic',
                            unique=True,
                            mysql_length={'namespace': 200,
                                          'mnemonic': 10},
                            ),
                      )

    # column definitions
    namespace = Column('namespace', VARCHAR(length=2048), nullable=False)
    mnemonic = Column('mnemonic', VARCHAR(length=2048), nullable=False)
    fullname = Column('fullname', VARCHAR(length=2048))
    cusip = Column('cusip', VARCHAR(length=2048))
    fraction = Column('fraction', INTEGER(), nullable=False)
    quote_flag = Column('quote_flag', INTEGER(), nullable=False)
    quote_source = Column('quote_source', VARCHAR(length=2048))
    quote_tz = Column('quote_tz', VARCHAR(length=2048))

    @property
    def base_currency(self):
        b = self.book
        if b is None:
            raise GnucashException("The commodity should be linked to a session to have a 'base_currency'")

        if self.namespace == "CURRENCY":
            # get the base currency as first commodity in DB
            return b.default_currency
        else:
            # retrieve currency from quoted_currency kvp
            # TODO: recover from the web (as fallback)
            mnemonic = self.get("quoted_currency", None)
            if mnemonic:
                return b.currencies(mnemonic=mnemonic)
            else:
                raise GnucashException("The commodity '{}' has no information about its base currency. "
                                       "Add a kvp item named 'quoted_currency' with the mnemonic of the "
                                       "currency to have proper behavior".format(self.mnemonic))

    # relation definitions
    accounts = relation('Account',
                        back_populates='commodity',
                        cascade='all, delete-orphan',
                        collection_class=CallableList)
    transactions = relation('Transaction',
                            back_populates='currency',
                            cascade='all, delete-orphan',
                            collection_class=CallableList,
                            )
    prices = relation("Price",
                      back_populates='commodity',
                      foreign_keys=[Price.commodity_guid],
                      cascade='all, delete-orphan',
                      lazy="dynamic",
                      )

    def __init__(self,
                 namespace,
                 mnemonic,
                 fullname,
                 fraction=100,
                 cusip="",
                 quote_flag=0,
                 quote_source=None,
                 quote_tz='',
                 book=None):

        if quote_source is None:
            quote_source = "currency" if namespace == "CURRENCY" else "yahoo"

        if book is not None:
            book.add(self)

        self.namespace = namespace
        self.mnemonic = mnemonic
        self.fullname = fullname
        self.fraction = fraction
        self.cusip = cusip
        self.quote_flag = quote_flag
        self.quote_source = quote_source
        self.quote_tz = quote_tz

        if book is not None:
            book.flush()

    def __str__(self):
        return "Commodity<{}:{}>".format(self.namespace, self.mnemonic)

    @property
    def precision(self):
        return len(str(self.fraction)) - 1

    def currency_conversion(self, currency):
        """
        Return the latest conversion factor to convert self to currency

        Attributes:
            currency (:class:`piecash.core.commodity.Commodity`): the currency to which the Price need to be converted

        Returns:
            a Decimal that can be multiplied by an amount expressed in self.commodity to get an amount expressed in currency

        Raises:
            GncConversionError: not possible to convert self to the currency

        """
        # conversion is done from self.commodity to commodity (if possible)
        sc2c = self.prices.filter_by(currency=currency).order_by(Price.date.desc()).first()
        if sc2c:
            return sc2c.value

        # conversion is done directly from commodity to self.commodity (if possible)
        c2sc = currency.prices.filter_by(currency=self).order_by(Price.date.desc()).first()
        if c2sc:
            return Decimal(1) / c2sc.value

        raise GncConversionError("Cannot convert {} to {}".format(self, currency))

    def update_prices(self, start_date=None):
        """
        Retrieve online prices for the commodity:

        - for currencies, it will get from quandl the exchange rates between the currency and its base_currency
        - for stocks, it will get from yahoo the daily closing prices expressed in its base_currency

        Args:
            start_date (:class:`datetime.date`): prices will be updated as of the start_date. If None, start_date is today
            - 7 days.

        .. note:: if prices are already available in the GnuCash file, the function will only retrieve prices as of the
           max(start_date, last quoted price date)

        .. todo:: add some frequency to retrieve prices only every X (week, month, ...)
        """
        if self.book is None:
            raise GncPriceError("Cannot update price for a commodity not attached to a book")

        # get last_price updated
        last_price = self.prices.order_by(Price.date.desc()).limit(1).first()

        if start_date is None:
            start_date = datetime.datetime.today().date() + datetime.timedelta(days=-7)

        if last_price:
            start_date = max(last_price.date + datetime.timedelta(days=1),
                             start_date)

        if self.namespace == "CURRENCY":
            # get reference currency (from book.root_account)
            default_currency = self.base_currency
            if default_currency == self:
                raise GncPriceError("Cannot update exchange rate for base currency")

            # through Quandl for exchange rates
            quotes = quandl_fx(self.mnemonic, default_currency.mnemonic, start_date)
            for q in quotes:
                p = Price(commodity=self,
                          currency=default_currency,
                          date=datetime.datetime.strptime(q.date, "%Y-%m-%d").date(),
                          value=str(q.rate))

        else:
            symbol = self.mnemonic
            share = get_latest_quote(symbol)
            currency = self.book.currencies(mnemonic=share.currency)
            tz = pytz.timezone(share.timezone)

            # get historical data
            for q in download_quote(
                    symbol,
                    start_date,
                    datetime.date.today(),
                    tz
            ):
                Price(commodity=self,
                      currency=currency,
                      date=q.date,
                      value=q.close,
                      type='last')

    def object_to_validate(self, change):
        if change[-1] != "deleted":
            yield self

    def validate(self):
        # check uniqueness of namespace/mnemonic
        try:
            self.book.query(Commodity).filter_by(namespace=self.namespace, mnemonic=self.mnemonic).one()
        except MultipleResultsFound:
            raise ValueError("{} already exists in this book".format(self))



from __future__ import division, unicode_literals
from __future__ import print_function

import datetime
import logging
import sys
import unicodedata

import pytz
import tzlocal
from sqlalchemy import types, Table, MetaData, ForeignKeyConstraint, event, create_engine
from sqlalchemy.dialects import sqlite
from sqlalchemy.ext.compiler import compiles
from sqlalchemy.ext.declarative import as_declarative
from sqlalchemy.ext.hybrid import hybrid_property
from sqlalchemy.orm import exc as orm_exc
from sqlalchemy.orm import sessionmaker, object_session

# import yaml


def __init__blocked(self, *args, **kwargs):
    raise NotImplementedError("Objects of type {} cannot be created from scratch "
                              "(only read)".format(self.__class__.__name__))


@as_declarative(constructor=__init__blocked)
class DeclarativeBase(object):
    @property
    def book(self):
        """Return the gnc book holding the object
        """
        s = object_session(self)
        return s and s.book

    def object_to_validate(self, change):
        """yield the objects to validate when the object is modified (change="new" "deleted" or "dirty").

        For instance, if the object is a Split, if it changes, we want to revalidate not the split
        but its transaction and its lot (if any). split.object_to_validate should yeild both split.transaction
        and split.lot
        """
        return
        yield

    def on_book_add(self):
        """Call when the object is added to a book"""
        pass

    def validate(self):
        """This must be reimplemented for object requiring validation
        """
        raise NotImplementedError(self)

    def get_all_changes(self):
        try:
            return self.book.session._all_changes[id(self)]
        except KeyError:
            return {"STATE_CHANGES": ["unchanged"],
                    "OBJECT": self}

    def __repr__(self):
        return str(self)


tz = tzlocal.get_localzone()
utc = pytz.utc


@compiles(sqlite.DATE, 'sqlite')
def compile_date(element, compiler, **kw):
    return "TEXT(8)"  # % element.__class__.__name__


@compiles(sqlite.DATETIME, 'sqlite')
def compile_datetime(element, compiler, **kw):
    """ data type for the date field

    note: it went from TEXT(14) in 2.6 to TEXT(19) in 2.8 to accommodate
    for the new ISO format of date in sqlite"""
    return "TEXT(14)"


class _DateTime(types.TypeDecorator):
    """Used to customise the DateTime type for sqlite (ie without the separators as in gnucash
    """
    impl = types.TypeEngine

    def load_dialect_impl(self, dialect):
        if dialect.name == "sqlite":
            return sqlite.DATETIME(
                storage_format="%(year)04d-%(month)02d-%(day)02d %(hour)02d:%(minute)02d:%(second)02d",
                regexp=r"(\d{4})-?(\d{2})-?(\d{2}) ?(\d{2}):?(\d{2}):?(\d{2})",
            )
        else:
            return types.DateTime()

    def process_bind_param(self, value, dialect):
        if value is not None:
            assert isinstance(value, datetime.datetime), "value {} is not of type datetime.datetime but type {}".format(
                value, type(value))
            if value.microsecond != 0:
                logging.warning("A datetime has been given with microseconds which are not saved in the database")

            if not value.tzinfo:
                value = tz.localize(value)

            return value.astimezone(utc).replace(tzinfo=None)

    def process_result_value(self, value, dialect):
        if value is not None:
            return utc.localize(value).astimezone(tz)


class _DateAsDateTime(types.TypeDecorator):
    """Used to customise the DateTime type for sqlite (ie without the separators as in gnucash
    """
    impl = types.TypeEngine

    def __init__(self, neutral_time=True, *args, **kwargs):
        super(_DateAsDateTime, self).__init__(*args, **kwargs)
        self.neutral_time = neutral_time

    def load_dialect_impl(self, dialect):
        if dialect.name == "sqlite":
            return sqlite.DATETIME(
                storage_format="%(year)04d-%(month)02d-%(day)02d %(hour)02d:%(minute)02d:%(second)02d",
                regexp=r"(\d{4})-?(\d{2})-?(\d{2}) ?(\d{2}):?(\d{2}):?(\d{2})",
            )
        else:
            return types.DateTime()

    def process_bind_param(self, value, dialect):
        if value is not None:
            assert isinstance(value, datetime.date) and not isinstance(value, datetime.datetime), \
                "value {} is not of type datetime.date but type {}".format(value, type(value))
            if self.neutral_time:
                result = datetime.datetime.combine(value, datetime.time(10, 59, 0))
            else:
                result = tz.localize(datetime.datetime.combine(value, datetime.time(0, 0, 0))) \
                    .astimezone(utc)
            return result.replace(tzinfo=None)

    def process_result_value(self, value, dialect):
        if value is not None:
            r = utc.localize(value).astimezone(tz).date()
            return r


class _Date(types.TypeDecorator):
    """Used to customise the DateTime type for sqlite (ie without the separators as in gnucash
    """
    impl = types.TypeEngine
    is_sqlite = False

    def load_dialect_impl(self, dialect):
        if dialect.name == "sqlite":
            return sqlite.DATE(
                storage_format="%(year)04d%(month)02d%(day)02d",
                regexp=r"(\d{4})(\d{2})(\d{2})"
            )
        else:
            return types.Date()


def mapped_to_slot_property(col, slot_name, slot_transform=lambda x: x):
    """Assume the attribute in the class as the same name as the table column with "_" prepended"""
    col_name = "_{}".format(col.name)

    def fget(self):
        return getattr(self, col_name)

    def fset(self, value):
        v = slot_transform(value)
        if v is None:
            if slot_name in self:
                del self[slot_name]
        else:
            self[slot_name] = v

        setattr(self, col_name, value)

    def expr(cls):
        return col

    return hybrid_property(
        fget=fget,
        fset=fset,
        expr=expr,
    )


def pure_slot_property(slot_name, slot_transform=lambda x: x,
                       ignore_invalid_slot=False):
    """
    Create a property (class must have slots) that maps to a slot

    :param slot_name: name of the slot
    :param slot_transform: transformation to operate before assigning value
    :param ignore_invalid_slot: True if incorrect values (usually due to deleted data)
        should be converted to None
    :return:
    """

    def fget(self):
        # return None if the slot does not exist. alternative could be to raise an exception
        try:
            return self[slot_name].value
        except KeyError:
            return None
        except orm_exc.NoResultFound:
            if ignore_invalid_slot:
                return None
            else:
                raise

    def fset(self, value):
        v = slot_transform(value)
        if v is None:
            if slot_name in self:
                del self[slot_name]
        else:
            self[slot_name] = v

    return hybrid_property(
        fget=fget,
        fset=fset,
    )


def kvp_attribute(name, to_gnc=lambda v: v, from_gnc=lambda v: v, default=None):
    def getter(self):
        try:
            return from_gnc(self[name].value)
        except KeyError:
            return default

    def setter(self, value):
        if value == default:
            try:
                del self[name]
            except KeyError:
                pass
        else:
            self[name] = to_gnc(value)

    return property(getter, setter)


def get_foreign_keys(metadata, engine):
    """ Retrieve all foreign keys from metadata bound to an engine
    :param metadata:
    :param engine:
    :return:
    """
    reflected_metadata = MetaData()
    for table_name in list(metadata.tables.keys()):
        table = Table(
            table_name,
            reflected_metadata,
            autoload=True,
            autoload_with=engine,
        )

        for constraint in table.constraints:
            if not isinstance(constraint, ForeignKeyConstraint):
                continue
            yield constraint


Session = sessionmaker(autoflush=False)


def create_piecash_engine(uri_conn, **kwargs):
    eng = create_engine(uri_conn, **kwargs)

    if eng.name == "sqlite":
        # add proper isolation code for sqlite engine
        @event.listens_for(eng, "connect")
        def do_connect(dbapi_connection, connection_record):
            # disable pysqlite's emitting of the BEGIN statement entirely.
            # also stops it from emitting COMMIT before any DDL.
            # print("=========================== in DO CONNECT")
            # dbapi_connection.isolation_level = "IMMEDIATE"
            # dbapi_connection.isolation_level = "EXCLUSIVE"
            pass

        @event.listens_for(eng, "begin")
        def do_begin(conn):
            # emit our own BEGIN
            # print("=========================== in DO BEGIN")
            # conn.execute("BEGIN EXCLUSIVE")
            pass

    return eng


class ChoiceType(types.TypeDecorator):
    impl = types.INTEGER()

    def __init__(self, choices, **kw):
        self.choices = dict(choices)
        super(ChoiceType, self).__init__(**kw)

    def process_bind_param(self, value, dialect):
        try:
            return [k for k, v in self.choices.items() if v == value][0]
        except IndexError:
            # print("Value '{}' is not in [{}]".format(", ".join(self.choices.values())))
            raise ValueError("Value '{}' is not in choices [{}]".format(value, ", ".join(self.choices.values())))

    def process_result_value(self, value, dialect):
        return self.choices[value]



# coding=utf-8
from __future__ import print_function

import io
import os.path

import jinja2

from piecash import open_book

if __name__ == '__main__':
    this_folder = os.path.dirname(os.path.realpath(__file__))

    with open_book(os.path.join(this_folder, "..", "gnucash_books", "CGT2015.gnucash"), open_if_lock=True) as book:
        transactions = book.transactions

        env = jinja2.Environment(trim_blocks=True, lstrip_blocks=True)
        xml = env.from_string(u"""
<?xml version="1.0"?>
<comptabilite>
  <exercice>
    <DateCloture>2016-12-31T00:00:00</DateCloture>
    <journal>
      <JournalCode>Le code du Journal</JournalCode>
      <JournalLib>Le libellé du Journal</JournalLib>
      {% for i, ecriture in enumerate(transactions) %}
      <ecriture>
        <EcritureNum>{{ i }}</EcritureNum>
        <EcritureDate>{{ ecriture.post_date.strftime("%Y-%m-%d") }}</EcritureDate>
        <EcritureLib>{{ ecriture.description }}</EcritureLib>
        <PieceRef>{{ ecriture.num }}</PieceRef>
        <PieceDate>{{ ecriture.post_date.strftime("%Y-%m-%d") }}</PieceDate>
        <ValidDate>{{ ecriture.post_date.strftime("%Y-%m-%d") }}</ValidDate>
        {% for sp in ecriture.splits %}
        <ligne>
          <CompteNum>{{ sp.account.code }}</CompteNum>
          <CompteLib>{{ sp.account.name }}</CompteLib>
          <CompteAuxNum>Le numéro de compte auxiliaire (à blanc si non utilisé)</CompteAuxNum>
          <CompteAuxLib>Le libellé de compte auxiliaire (à blanc si non utilisé)</CompteAuxLib>
          <Montantdevise></Montantdevise>
          <Montant>{{ abs(sp.value) }}</Montant>
          <Sens>{% if sp.value >0 %}c{% else %}d{% endif %}</Sens>
        </ligne>
        {% endfor %}
      </ecriture>
      {% endfor %}
    </journal>
  </exercice>
</comptabilite>
        """).render(transactions=transactions,
                    enumerate=enumerate,
                    abs=abs,
                    )

        with io.open("resultat.xml", "w", encoding="utf-8") as f:
            f.write(xml)

# pour référence, fichier généré à partir du xsd
"""
<?xml version="1.0"?>
<comptabilite>
  <exercice>
    <DateCloture>2007-10-26T08:36:28</DateCloture>
    <!--1 or more repetitions:-->
    <journal>
      <JournalCode>string</JournalCode>
      <JournalLib>string</JournalLib>
      <!--Zero or more repetitions:-->
      <ecriture>
        <EcritureNum>string</EcritureNum>
        <EcritureDate>2014-06-09+02:00</EcritureDate>
        <EcritureLib>string</EcritureLib>
        <PieceRef>string</PieceRef>
        <PieceDate>2009-05-16T14:42:28</PieceDate>
        <!--Optional:-->
        <EcritureLet>string</EcritureLet>
        <!--Optional:-->
        <DateLet>2002-11-05T09:01:03+01:00</DateLet>
        <ValidDate>2016-01-01T20:07:42</ValidDate>
        <!--2 or more repetitions:-->
        <ligne>
          <CompteNum>string</CompteNum>
          <CompteLib>string</CompteLib>
          <!--You have a CHOICE of the next 2 items at this level-->
          <!--Optional:-->
          <CompAuxNum>string</CompAuxNum>
          <!--Optional:-->
          <CompteAuxNum>string</CompteAuxNum>
          <!--You have a CHOICE of the next 2 items at this level-->
          <!--Optional:-->
          <CompAuxLib>string</CompAuxLib>
          <!--Optional:-->
          <CompteAuxLib>string</CompteAuxLib>
          <!--Optional:-->
          <Montantdevise>string</Montantdevise>
          <!--Optional:-->
          <Idevise>string</Idevise>
          <!--You have a CHOICE of the next 3 items at this level-->
          <Debit>1.5E2</Debit>
          <Credit>1.5E2</Credit>
          <Montant>1.5E2</Montant>
          <Sens>c</Sens>
          <!--You may enter ANY elements at this point-->
          <AnyElement/>
        </ligne>
        <ligne>
          <CompteNum>string</CompteNum>
          <CompteLib>string</CompteLib>
          <!--You have a CHOICE of the next 2 items at this level-->
          <!--Optional:-->
          <CompAuxNum>string</CompAuxNum>
          <!--Optional:-->
          <CompteAuxNum>string</CompteAuxNum>
          <!--You have a CHOICE of the next 2 items at this level-->
          <!--Optional:-->
          <CompAuxLib>string</CompAuxLib>
          <!--Optional:-->
          <CompteAuxLib>string</CompteAuxLib>
          <!--Optional:-->
          <Montantdevise>string</Montantdevise>
          <!--Optional:-->
          <Idevise>string</Idevise>
          <!--You have a CHOICE of the next 3 items at this level-->
          <Debit>1.5E2</Debit>
          <Credit>1.5E2</Credit>
          <Montant>1.5E2</Montant>
          <Sens>c</Sens>
          <!--You may enter ANY elements at this point-->
          <AnyElement/>
        </ligne>
      </ecriture>
    </journal>
  </exercice>
</comptabilite>
"""



import datetime
import uuid
from collections import defaultdict
from decimal import Decimal

from sqlalchemy import Column, VARCHAR, ForeignKey, BIGINT, INTEGER, Index
from sqlalchemy.orm import relation, validates, foreign
from sqlalchemy.orm.base import NEVER_SET

from .._common import CallableList, GncImbalanceError
from .._common import GncValidationError, hybrid_property_gncnumeric, Recurrence
from .._declbase import DeclarativeBaseGuid
from ..sa_extra import _Date, _DateTime, mapped_to_slot_property, pure_slot_property, _DateAsDateTime


class Split(DeclarativeBaseGuid):
    """
    A GnuCash Split.

    .. note::

        A split used in a scheduled transaction has its main attributes in form of slots.

    Attributes:
        transaction(:class:`piecash.core.transaction.Transaction`): transaction of the split
        account(:class:`piecash.core.account.Account`): account of the split
        lot(:class:`piecash.business.Lot`): lot to which the split pertains
        memo(str): memo of the split
        value(:class:`decimal.Decimal`): amount express in the currency of the transaction of the split
        quantity(:class:`decimal.Decimal`): amount express in the commodity of the account of the split
        reconcile_state(str): 'n', 'c' or 'y'
        reconcile_date(:class:`datetime.datetime`): time
        action(str): describe the type of action behind the split (free form string but with dropdown in the GUI
    """
    __tablename__ = 'splits'

    __table_args__ = (
        # indices
        Index('splits_tx_guid_index', 'tx_guid'),
        Index('splits_account_guid_index', 'account_guid'),
    )

    # column definitions
    # the transaction_guid is not mandatory at construction time because it can be set through a tr.splits.append(...) operation
    # however, in the validation of the object, we raise an error if there is no transaction set at that time
    transaction_guid = Column('tx_guid', VARCHAR(length=32), ForeignKey('transactions.guid'))
    account_guid = Column('account_guid', VARCHAR(length=32), ForeignKey('accounts.guid'), nullable=False)
    memo = Column('memo', VARCHAR(length=2048), nullable=False)
    action = Column('action', VARCHAR(length=2048), nullable=False)

    reconcile_state = Column('reconcile_state', VARCHAR(length=1), nullable=False)
    reconcile_date = Column('reconcile_date', _DateTime())

    _value_num = Column('value_num', BIGINT(), nullable=False)
    _value_denom = Column('value_denom', BIGINT(), nullable=False)
    _value_denom_basis = None
    value = hybrid_property_gncnumeric(_value_num, _value_denom)
    _quantity_num = Column('quantity_num', BIGINT(), nullable=False)
    _quantity_denom = Column('quantity_denom', BIGINT(), nullable=False)
    _quantity_denom_basis = None
    quantity = hybrid_property_gncnumeric(_quantity_num, _quantity_denom)

    lot_guid = Column('lot_guid', VARCHAR(length=32), ForeignKey('lots.guid'))

    # relation definitions
    account = relation('Account', back_populates='splits')
    lot = relation('Lot', back_populates='splits')
    transaction = relation('Transaction', back_populates='splits')

    @property
    def is_credit(self):
        return self.value < 0

    @property
    def is_debit(self):
        return self.value > 0

    def __init__(self,
                 account,
                 value,
                 quantity=None,
                 transaction=None,
                 memo="",
                 action="",
                 reconcile_date=None,
                 reconcile_state="n",
                 lot=None,
                 ):
        self.transaction = transaction
        self.account = account
        self.value = value
        self.quantity = value if quantity is None else quantity
        self.memo = memo
        self.action = action
        self.reconcile_date = reconcile_date
        self.reconcile_state = reconcile_state
        self.lot = lot

    def __str__(self):
        try:
            cur = self.transaction.currency.mnemonic
            acc = self.account
            com = acc.commodity.mnemonic
            if com == "template":
                # case of template split from scheduled transaction
                sched_xaction = self["sched-xaction"]
                credit = sched_xaction["credit-formula"].value
                debit = sched_xaction["debit-formula"].value
                return "SplitTemplate<{} {} {}>".format(sched_xaction["account"].value,
                                                        "credit={}".format(credit) if credit else "",
                                                        "debit={}".format(debit) if debit else "",

                                                        )
            elif cur == com:
                # case of same currency split
                return "Split<{} {} {}>".format(acc,
                                                self.value, cur)
            else:
                # case of non currency split
                return "Split<{} {} {} [{} {}]>".format(acc,
                                                        self.value, cur,
                                                        self.quantity, com)
        except AttributeError:
            return "Split<{}>".format(self.account)

    def object_to_validate(self, change):
        yield self
        if self.transaction:
            yield self.transaction
        if self.lot:
            yield self.lot

    def validate(self):
        old = self.get_all_changes()

        if old["STATE_CHANGES"][-1] == "deleted":
            return

        if '_quantity_num' in old or '_value_num' in old:
            self.transaction._recalculate_balance = True

        if self.transaction_guid is None:
            raise GncValidationError("The split is not linked to a transaction")

        if self.transaction.currency == self.account.commodity:
            if self.quantity != self.value:
                raise GncValidationError("The split has a quantity diffeerent from value "
                                         "while the transaction currency and the account commodity is the same")
        else:
            if self.quantity is None:
                raise GncValidationError(
                    "The split quantity is not defined while the split is on a commodity different from the transaction")
            # Allow for either value to be 0.0 (or -0.0).
            if self.quantity * self.value < 0:
                raise GncValidationError("The split quantity has not the same sign as the split value")

        # everything is fine, let us normalise the value with respect to the currency/commodity precisions
        self._quantity_denom_basis = self.account.commodity_scu
        self._value_denom_basis = self.transaction.currency.fraction

        if self.transaction.currency != self.account.commodity and self.quantity:
            # let us also add a Price
            from piecash import Price

            value = (self.value / self.quantity).quantize(Decimal("0.000001"))
            try:
                # find existing price if any and if so, do nothing
                pr = self.book.prices(commodity=self.account.commodity,
                                      currency=self.transaction.currency,
                                      date=self.transaction.post_date,
                                      )

            except KeyError:
                # otherwise, add a price in the database
                pr = Price(commodity=self.account.commodity,
                           currency=self.transaction.currency,
                           date=self.transaction.post_date,
                           value=value,
                           type="transaction",
                           source="user:split-register")

            # and an action if not yet defined
            if self.action == "":
                self.action = "Sell" if self.quantity.is_signed() else "Buy"


class Transaction(DeclarativeBaseGuid):
    """
    A GnuCash Transaction.

    Attributes:
        currency (:class:`piecash.core.commodity.Commodity`): currency of the transaction. This attribute is
            write-once (i.e. one cannot change it after being set)
        description (str): description of the transaction
        enter_date (:class:`datetime.datetime`): datetimetime at which transaction is entered
        post_date (:class:`datetime.date`): day on which transaction is posted
        num (str): user provided transaction number
        splits (list of :class:`Split`): list of the splits of the transaction
        scheduled_transaction  (:class:`ScheduledTransaction`): scheduled transaction behind the transaction
        notes (str): notes on the transaction (provided via a slot)
    """
    __tablename__ = 'transactions'

    __table_args__ = (
        Index('tx_post_date_index', 'post_date'),
    )

    # column definitions
    currency_guid = Column('currency_guid', VARCHAR(length=32), ForeignKey('commodities.guid'), nullable=False)
    num = Column('num', VARCHAR(length=2048), nullable=False)
    _post_date = Column('post_date', _DateAsDateTime(neutral_time=True))
    post_date = mapped_to_slot_property(_post_date,
                                        slot_name="date-posted",
                                        # slot_transform=lambda x: x.date() if x else None
                                        )
    enter_date = Column('enter_date', _DateTime)
    description = Column('description', VARCHAR(length=2048))
    notes = pure_slot_property('notes')

    scheduled_transaction = pure_slot_property('from-sched-xaction', ignore_invalid_slot=True)

    # relation definitions
    currency = relation('Commodity',
                        back_populates='transactions',
                        )
    splits = relation('Split',
                      back_populates="transaction",
                      cascade='all, delete-orphan',
                      collection_class=CallableList,
                      )

    def __init__(self,
                 currency,
                 description="",
                 notes=None,
                 splits=None,
                 enter_date=None,
                 post_date=None,
                 num="",
                 ):

        if not (enter_date is None or type(enter_date) is datetime.datetime):
            raise GncValidationError("enter_date should be a datetime object")
        if not (post_date is None or type(post_date) is datetime.date):
            raise GncValidationError("post_date should be a date object")

        self.currency = currency
        self.description = description
        self.enter_date = (enter_date
                           if enter_date
                           else datetime.datetime.now()).replace(microsecond=0)
        self.post_date = (post_date if post_date else datetime.date.today())
        self.num = num
        if notes is not None:
            self.notes = notes
        if splits:
            self.splits = splits

    def __str__(self):
        return "Transaction<[{}] '{}' on {:%Y-%m-%d}{}>".format(self.currency.mnemonic,
                                                                self.description,
                                                                self.post_date,
                                                                " (from sch tx)" if self.scheduled_transaction else "")

    def object_to_validate(self, change):
        yield self

    def validate(self):
        old = self.get_all_changes()

        if old["STATE_CHANGES"][-1] == "deleted":
            return

        if self.currency.namespace != "CURRENCY":
            raise GncValidationError("You are assigning a non currency commodity to a transaction")

        # check all accounts related to the splits of the transaction are not placeholder(=frozen)
        for sp in self.splits:
            if sp.account.placeholder != 0:
                raise GncValidationError("Account '{}' used in the transaction is a placeholder".format(sp.account))

        # check same currency
        if "currency" in old and old["currency"] is not NEVER_SET:
            raise GncValidationError("You cannot change the currency of a transaction once it has been set")

        # validate the splits
        if hasattr(self, "_recalculate_balance"):
            del self._recalculate_balance
            value_imbalance, quantity_imbalances = self.calculate_imbalances()
            if value_imbalance:
                # raise exception instead of creating an imbalance entry as probably an error
                # (in the gnucash GUI, another decision taken because need of "save unfinished transaction")
                raise GncImbalanceError("The transaction {} is not balanced on its value".format(self))

            if any(quantity_imbalances.values()) and self.book.use_trading_accounts:
                self.normalize_trading_accounts()

        # normalise post_date to 10:59AM
        # if self.post_date:
        #    self.post_date = self.post_date.replace(hour=10, minute=59, second=0, microsecond=0, tzinfo=utc)

    def calculate_imbalances(self):
        """Calculate value and quantity imbalances of a transaction"""
        value_imbalance = Decimal(0)  # hold imbalance on split.value
        quantity_imbalances = defaultdict(Decimal)  # hold imbalance on split.quantity per cdty

        # collect imbalance information
        for sp in self.splits:
            value_imbalance += sp.value
            quantity_imbalances[sp.account.commodity] += sp.quantity

        return value_imbalance, quantity_imbalances

    def normalize_trading_accounts(self):
        # collect imbalance information
        classic_splits = defaultdict(list)
        trading_splits = defaultdict(list)
        trading_target_value = defaultdict(Decimal)
        trading_target_quantity = defaultdict(Decimal)

        for sp in self.splits:
            cdty = sp.account.commodity

            if sp.account.type == "TRADING":
                trading_splits[cdty].append(sp)
            else:
                classic_splits[cdty].append(sp)
            trading_target_value[cdty] += sp.value
            trading_target_quantity[cdty] += sp.quantity

        root = self.book.root_account
        # imbalance in quantities to be settled using trading accounts
        for cdty, v in trading_target_value.items():
            q = trading_target_quantity[cdty]

            # if commodity is balanced, do not do anything
            if (v == q == 0): continue

            # otherwise, look if there is some trading imbalance (ie a split with the trading account already exists!)
            if cdty in trading_splits:
                # and adjust the related split to rebalance
                sp, = trading_splits[cdty]
                sp.value -= v
                sp.quantity -= q
            else:
                # otherwise, we must create the split related to the trading account
                # assume trading account exists
                t_acc = self.book.trading_account(cdty)
                sp = Split(account=t_acc,
                           value=-v,
                           quantity=-q,
                           transaction=self,
                           )


class ScheduledTransaction(DeclarativeBaseGuid):
    """
    A GnuCash Scheduled Transaction.

    Attributes
        adv_creation (int) : days to create in advance (0 if disabled)
        adv_notify (int) : days to notify in advance (0 if disabled)
        auto_create (bool) :
        auto_notify (bool) :
        enabled (bool) :
        start_date (:class:`datetime.datetime`) : date to start the scheduled transaction
        last_occur (:class:`datetime.datetime`) : date of last occurence of the schedule transaction
        end_date (:class:`datetime.datetime`) : date to end the scheduled transaction (num/rem_occur should be 0)
        instance_count (int) :
        name (str) : name of the scheduled transaction
        num_occur (int) : number of occurences in total (end_date should be null)
        rem_occur (int) : number of remaining occurences (end_date should be null)
        template_account (:class:`piecash.core.account.Account`): template account of the transaction
    """
    __tablename__ = 'schedxactions'

    __table_args__ = {}

    # column definitions
    guid = Column('guid', VARCHAR(length=32), primary_key=True, nullable=False, default=lambda: uuid.uuid4().hex)
    name = Column('name', VARCHAR(length=2048))
    enabled = Column('enabled', INTEGER(), nullable=False)
    start_date = Column('start_date', _Date())
    end_date = Column('end_date', _Date())
    last_occur = Column('last_occur', _Date())
    num_occur = Column('num_occur', INTEGER(), nullable=False)
    rem_occur = Column('rem_occur', INTEGER(), nullable=False)
    auto_create = Column('auto_create', INTEGER(), nullable=False)
    auto_notify = Column('auto_notify', INTEGER(), nullable=False)
    adv_creation = Column('adv_creation', INTEGER(), nullable=False)
    adv_notify = Column('adv_notify', INTEGER(), nullable=False)
    instance_count = Column('instance_count', INTEGER(), nullable=False)
    template_act_guid = Column('template_act_guid', VARCHAR(length=32), ForeignKey('accounts.guid'), nullable=False)

    # relation definitions
    template_account = relation('Account')
    recurrence = relation('Recurrence',
                          primaryjoin=guid == foreign(Recurrence.obj_guid),
                          cascade='all, delete-orphan',
                          uselist=False,
                          )

    def __str__(self):
        return "ScheduledTransaction<'{}' {}>".format(self.name, self.recurrence)


class Lot(DeclarativeBaseGuid):
    """
    A GnuCash Lot. Each lot is linked to an account. Splits in this account can be associated to a Lot. Whenever
    the balance of the splits goes to 0, the Lot is closed (otherwise it is opened)

    Attributes:
        is_closed (int) : 1 if lot is closed, 0 otherwise
        account (:class:`piecash.core.account.Account`): account of the Lot
        splits (:class:`piecash.core.transaction.Split`): splits associated to the Lot
    """
    __tablename__ = 'lots'

    __table_args__ = {}

    # column definitions
    account_guid = Column('account_guid', VARCHAR(length=32), ForeignKey('accounts.guid'))
    is_closed = Column('is_closed', INTEGER(), nullable=False)

    title = pure_slot_property('title')
    notes = pure_slot_property('notes')

    # relation definitions
    account = relation('Account', back_populates='lots', )
    splits = relation('Split',
                      back_populates='lot',
                      collection_class=CallableList,
                      )

    def __init__(self,
                 title,
                 account,
                 notes="",
                 splits=None,
                 is_closed=0):
        self.title = title
        self.account = account
        self.notes = notes
        if splits:
            self.splits[:] = splits
        self.is_closed = is_closed

    @validates("splits", "account")
    def check_no_change_if_lot_is_close(self, key, value):
        if self.is_closed:
            raise ValueError("Lot is closed and cannot be changed (adding splits or changing account")
        return value

    def object_to_validate(self, change):
        yield self

    def validate(self):
        # check all splits have same account
        for sp in self.splits:
            if sp.account != self.account:
                raise ValueError("Split {} is not in the same commodity of the lot {}".format(sp, self))

    def __str__(self):
        return "Lot<'{}' on {}>".format(self.title, self.account.name)



# -*- coding: utf-8 -*-
"""Python interface to GnuCash documents"""
from . import metadata

__version__ = metadata.version
__author__ = metadata.authors[0]
__license__ = metadata.license
__copyright__ = metadata.copyright

from ._common import (
    GncNoActiveSession,
    GnucashException, GncValidationError, GncImbalanceError,
    Recurrence
)
from .core import (
    Book,
    Account, ACCOUNT_TYPES, AccountType,
    Transaction, Split, ScheduledTransaction, Lot,
    Commodity, Price,
    create_book, open_book,
    factories,
)
from .business import Vendor, Customer, Employee, Address
from .business import Invoice, Job
from .business import Taxtable, TaxtableEntry
from .budget import Budget, BudgetAmount
from .kvp import slot
from .ledger import ledger



#!/usr/local/bin/python
"""original script from https://github.com/MatzeB/pygnucash/blob/master/gnucash2ledger.py by Matthias Braun matze@braunis.de
 adapted for:
 - python 3 support
 - new string formatting

The security must exist in GnuCash database.
Example of a valid CSV file:
    currency,commodity,value,date
    AUD,YMAX,9.06,"2017-11-10"
 
"""
import argparse
import csv
import sys
from datetime import datetime
from decimal import Decimal

import piecash
from piecash import Price

parser = argparse.ArgumentParser(description="""
Import and export prices to and from a gnucash book.

Per default, it exports the full prices list from the gnucash book to the standard output in a CSV format.
To import, add the argument "--import file_to_import.csv".

The format used is a standard CSV (comma as separator) with the following columns:
 - date (YYYY-MM-DD)
 - commodity (gnucash mnemonic)
 - currency (gnucash mnemonic)
 - type (string, optional)
 - value (float)
 - type (string, optional)
""", formatter_class=argparse.RawTextHelpFormatter)
parser.add_argument("gnucash_filename",
                    help="the name of the gnucash file")
parser.add_argument("--import", dest="operation",
                    help="to import the prices from a csv file (default export)")
args = parser.parse_args()

# args.operation is the name of the import file (.csv).

if args.operation is None:
    # export the prices
    sys.stdout.write("date,type,value,value_num, value_denom, currency,commodity,source\n")
    with piecash.open_book(args.gnucash_filename, open_if_lock=True) as book:
        sys.stdout.writelines(
            "{p.date:%Y-%m-%d},{p.type},{p.value},{p._value_num},{p._value_denom},{p.currency.mnemonic},{p.commodity.mnemonic},{p.source}\n".format(
                p=p) for p in book.prices)
else:
    # import the prices
    with piecash.open_book(args.gnucash_filename, open_if_lock=True, readonly=False) as book:
        cdty = book.commodities
        importFile = open(args.operation, 'r')

        for l in csv.DictReader(importFile):
            cur = cdty(mnemonic=l['currency'])
            com = cdty(mnemonic=l['commodity'])
            type = l.get('type', None)
            date = datetime.strptime(l['date'], "%Y-%m-%d")
            v = Decimal(l['value'])
            Price(currency=cur,
                  commodity=com,
                  date=date,
                  value=v,
                  source="piecash-importer")
        book.save()



from datetime import date

import piecash
from piecash import Commodity, Price

# create a new book
with piecash.create_book() as book:
    # create a commodity/stock
    cdty = Commodity("NYSE", "WOCA", "wonder cola", book=book)

    # create two quotes/prices
    p = Price(cdty, book.default_currency, date(2018, 9, 6), "3.44")
    p = Price(cdty, book.default_currency, date(2018, 9, 7), "3.52")

    # save the book
    book.save()

    # create a second price for the second day
    p = Price(cdty, book.default_currency, date(2018, 9, 7), "3.45")

    # save the book ==> raise exception when validating the book as duplicate price (same cdty x currency x day
    try:
        book.save()
        print("this code will not be reached")
    except ValueError:
        print("You cannot have two price in same currency for same commodity and same day")

    # cancel last operations
    book.cancel()

    # to change a price, retrieve it first and then modify it
    # retrieve price for given day
    price = book.prices(commodity=cdty, currency=book.default_currency, date=date(2018, 9, 7))
    print(price)
    price.value = "3.45"

    book.save()

    # check prices have been modified correctly
    print(book.prices)


    # you can also use a helper that will either add or modify a price
    def add_or_update_price(cdty, curr, day, value):
        try:
            price = book.prices(commodity=cdty, currency=curr, date=day)  # raise a KeyError if Price does not exist
            print("--> modifying existing price ", price)
            price.value = value
        except KeyError:
            price = Price(cdty, curr, day, value)
            print("--> adding new price ", price)


    add_or_update_price(cdty, book.default_currency, date(2018, 9, 9), 5)
    book.save()

    add_or_update_price(cdty, book.default_currency, date(2018, 9, 9), 6)
    book.save()



# coding=utf-8
from __future__ import unicode_literals

from decimal import Decimal

from test_helper import db_sqlite_uri, db_sqlite, new_book, new_book_USD, book_uri, book_invoices, Person
# dummy line to avoid removing unused symbols
from piecash import Address, Employee, Account, Invoice
from piecash.business import Taxtable, TaxtableEntry

a = db_sqlite_uri, db_sqlite, new_book, new_book_USD, book_uri, book_invoices, Person


class TestInvoice(object):
    """
    Person is a parameter taking values in [Customer, Vendor, Employee]
    """

    def test_create_person_noid_nobook(self, book_invoices):
        assert len(book_invoices.invoices) == 1
        invoice = book_invoices.invoices[0]
        assert invoice.charge_amt == 0
        assert len(invoice.entries) == 2
        entry = invoice.entries[0]
        assert entry.quantity == Decimal('25')
        assert entry.invoice == invoice



import sqlalchemy as sa
from sqlalchemy import create_engine, event, and_
from sqlalchemy.ext.declarative import as_declarative
from sqlalchemy.orm import sessionmaker, relationship, foreign, remote, backref
from sqlalchemy_utils import generic_relationship


@as_declarative()
class Base:
    pass


engine = create_engine("sqlite://")
session = sessionmaker(bind=engine)()


class HasAddresses(object):
    """HasAddresses mixin, creates a relationship to
    the address_association table for each parent.

    """


class User(HasAddresses, Base):
    __tablename__ = 'user'
    id = sa.Column(sa.Integer, primary_key=True)


class Customer(HasAddresses, Base):
    __tablename__ = 'customer'
    id = sa.Column(sa.Integer, primary_key=True)


class Event(Base):
    __tablename__ = 'event'
    id = sa.Column(sa.Integer, primary_key=True)

    # This is used to discriminate between the linked tables.
    object_type = sa.Column(sa.Unicode(255))

    # This is used to point to the primary key of the linked row.
    object_id = sa.Column(sa.Integer)

    @property
    def parent(self):
        """Provides in-Python access to the "parent" by choosing
        the appropriate relationship.

        """
        return getattr(self, "parent_%s" % self.object_type)

    @parent.setter
    def parent(self, value):
        value.__class__.__name__.lower()
        old_parent = self.parent
        self.object_type = value.__class__.__name__.lower()
        self.object_id = value.id
        session.expire(old_parent)


@event.listens_for(HasAddresses, "mapper_configured", propagate=True)
def setup_listener(mapper, class_):
    name = class_.__name__
    discriminator = name.lower()
    class_.object = relationship(Event,
                                 primaryjoin=and_(
                                     class_.id == foreign(remote(Event.object_id)),
                                     Event.object_type == discriminator
                                 ),
                                 backref=backref(
                                     "parent_%s" % discriminator,
                                     primaryjoin=and_(
                                         remote(class_.id) == foreign(Event.object_id),
                                         Event.object_type == discriminator
                                     )
                                 )
                                 )

    @event.listens_for(class_.object, "append")
    def append_address(target, value, initiator):
        value.object_type = discriminator


Base.metadata.create_all(engine)

if True:
    # Some general usage to attach an event to a user.
    user = User()
    for c in range(1):
        customer = Customer()
        session.add(customer)
    session.add_all([user])
    session.commit()

    ev = Event()
    user.object = [ev]
    ev1 = Event()
    customer.object = [ev1]
    # ev.parent = customer

    session.add(ev)
    session.commit()

if True:
    customer = session.query(Customer).all()[0]
    user = session.query(User).all()[0]
    ev = session.query(Event).all()[0]

# Find the event we just made.
print(ev.parent)

# ev.parent = customer
# print(ev.parent)
# session.commit()
# print(ev.parent)

# Find any events that are bound to users.
for ev in session.query(Event).all():
    print("-", ev, ev.parent)

for ev in session.query(Event).filter(Event.parent_user != None).all():
    print("o", ev, ev.parent)

for ev in session.query(Event).filter_by(parent_user=user).all():
    print("x", ev, ev.parent)
print(user.object)
print(customer.object)
print(ev, ev.parent)
ev.parent = customer
#session.commit()
print(user.object)
print(customer.object)
print(ev, ev.parent)



# coding=utf-8
from __future__ import unicode_literals

# from gnucash import Session, Account, Transaction, Split, GncNumeric
# import gnucash
from piecash import Account
from piecash import open_book


bookname = r"../gnucash_books/investment.gnucash"

with open_book(bookname,open_if_lock=True, echo=False) as b:
    veur = b.commodities(mnemonic="VEUR")
    eur = b.currencies(mnemonic="EUR")

    for acc in veur.accounts:
        print("{} account has a balance of {} and holds a quantity {} of {}".format(
            acc.name,
            # acc.currency,
            acc.get_balance(),
            acc.get_quantity(),
            acc.commodity))

    print("$"*100)
    for acc in eur.accounts:
        print("{} account has a balance of {} and holds a quantity {} of {}".format(
            acc.name,
            # acc.currency,
            acc.get_balance(),
            acc.get_quantity(),
            acc.commodity))



import pytest
from sqlalchemy import Column

from helper_slot_class import B
from piecash import Transaction
from piecash.kvp import DictWrapper, SlotInt, SlotGUID
from piecash.sa_extra import pure_slot_property, DeclarativeBase


class TestSlot_behavior(object):
    def test_pure_slot_property_int(self):
        class A(DictWrapper):
            slots = []
            prop = pure_slot_property('prop')

        a = A()
        assert len(a.slots) == 0
        assert a.prop is None

        # assignment of property
        a.prop = 4
        assert len(a.slots) == 1
        assert a.slots == [SlotInt(name="prop", value=4)]
        assert a.slots[0] != [SlotInt(name="prop", value=8)]
        assert a.slots[0] != [SlotInt(name="prop_notok", value=4)]
        assert a.prop == 4




#!/usr/local/bin/python
"""Basic script to export QIF. Heavily untested ..."""
import sys

# https://github.com/jemmyw/Qif/blob/master/QIF_references

import click

from piecash.scripts.cli import cli


@cli.command()
@click.argument('book', type=click.Path(exists=True))
@click.option('--output', type=click.File('w'), default="-",
              help="File to which to export the data (default=stdout)")
def qif(book, output):
    """Export to QIF format.

    This scripts export a GnuCash BOOK to the QIF format.
    """
    try:
        import qifparse.qif as _qif
    except ImportError:
        _qif = None
        print("You need to install the qifparse module ('pip install qifparse')")
        sys.exit()

    import piecash

    map_gnc2qif = {
        "CASH": 'Cash',
        "BANK": 'Bank',
        "RECEIVABLE": 'Bank',
        "PAYABLE": 'Ccard',
        "MUTUAL": 'Bank',
        "CREDIT": 'Ccard',
        "ASSET": 'Oth A',
        "LIABILITY": 'Oth L',
        "TRADING": 'Oth L',
        # 'Invoice',  # Quicken for business only
        "STOCK": 'Invst',
    }

    with piecash.open_book(book, open_if_lock=True) as s:
        b = _qif.Qif()
        map = {}
        for acc in s.accounts:
            if acc.parent == s.book.root_template: continue
            elif acc.type in ["INCOME", "EXPENSE", "EQUITY"]:
                item = _qif.Category(name=acc.fullname,
                                     description=acc.description,
                                     expense=acc.type == "EXPENSE",
                                     income=acc.type == "INCOME" or acc.type == "EQUITY",
                                     )
                b.add_category(item)
            elif acc.type in map_gnc2qif:
                actype = map_gnc2qif[acc.type]
                if actype=="Invst":
                    # take parent account
                    item = _qif.Account(name=acc.fullname, account_type=actype)
                else:
                    item = _qif.Account(name=acc.fullname, account_type=actype)
                b.add_account(item)
            else:
                print("unknow {} for {}".format(acc.type, acc.fullname))

            map[acc.fullname] = item

        # print(str(b))
        def split_type(sp):
            qif_obj = map[sp.account.fullname]
            if isinstance(qif_obj, _qif.Account):
                return qif_obj.account_type
            else:
                return "Expense" if qif_obj.expense else "Income"

        def sort_split(sp):
            type = split_type(sp)
            if type=="Invst":
                return 1
            elif type in ["Expense","Income"]:
                return 2
            else:
                return 0

        tpl = s.book.root_template

        for tr in s.transactions:
            if not tr.splits or len(tr.splits)<2: continue # split empty transactions
            if tr.splits[0].account.parent==tpl: continue # skip template transactions

            splits = sorted(tr.splits, key=sort_split)
            if all(sp.account.commodity.namespace == "CURRENCY" for sp in splits):

                sp1, sp2 = splits[:2]
                item = _qif.Transaction(date=tr.post_date,
                                        num=tr.num,
                                        payee=tr.description,
                                        amount=sp1.value,
                                        memo=sp1.memo,
                                        )
                if isinstance(map[sp2.account.fullname], _qif.Account):
                    item.to_account = sp2.account.fullname
                else:
                    item.category = sp2.account.fullname

                if len(splits) > 2:
                    for sp in splits[1:]:
                        if isinstance(map[sp.account.fullname], _qif.Account):
                            asp = _qif.AmountSplit(amount=-sp.value,
                                                   memo=sp.memo,
                                                   to_account=sp.account.fullname,
                                                   )
                        else:
                            asp = _qif.AmountSplit(amount=-sp.value,
                                                   memo=sp.memo,
                                                   category=sp.account.fullname,
                                                   )
                        item.splits.append(asp)
                map[sp1.account.fullname].add_transaction(item,
                                                          header="!Type:{}".format(map[sp1.account.fullname].account_type))
            else:
                # match pattern of splits for an investment

                sp_account, sp_security, sp_others = splits[0], splits[1], splits[2:]

                assert split_type(sp_account) in ["Bank", "Cash"]
                assert split_type(sp_security) in ["Invst"]
                assert all(split_type(sp)=="Expense" for sp in sp_others)
                assert sp_security.account.parent.type=="BANK", "Security account has no parent STOCK account (aka a Brokerage account)"
                item = _qif.Investment(date=tr.post_date,
                                       action="Buy" if sp_security.quantity>0 else "Sell",
                                       security=sp_security.account.commodity.mnemonic,
                                       price=sp_security.value / sp_security.quantity,
                                       quantity=sp_security.quantity,
                                       amount=sp_security.value,
                                       commission=sum(sp.value for sp in sp_others),
                                       first_line=tr.description,
                                       to_account=sp_account.account.fullname,
                                       amount_transfer=abs(sp_account.value),
                                       )
                map[sp_security.account.fullname]\
                    .add_transaction(item, header="!Type:{}".format(split_type(sp_security)))

    output.write(str(b))



from __future__ import division

import uuid

from sqlalchemy import Column, VARCHAR, INTEGER, BIGINT, ForeignKey
from sqlalchemy.orm import relation, foreign

from ._common import hybrid_property_gncnumeric, Recurrence, CallableList
from ._declbase import DeclarativeBaseGuid
from .sa_extra import DeclarativeBase


class Budget(DeclarativeBaseGuid):
    """
    A GnuCash Budget

    Attributes:
        name (str): name of the budget
        description (str): description of the budget
        amounts (list of :class:`piecash.budget.BudgetAmount`): list of amounts per account
    """
    __tablename__ = 'budgets'

    __table_args__ = {}

    # column definitions
    # keep this line as we reference it in the primaryjoin
    guid = Column('guid', VARCHAR(length=32), primary_key=True, nullable=False, default=lambda: uuid.uuid4().hex)
    name = Column('name', VARCHAR(length=2048), nullable=False)
    description = Column('description', VARCHAR(length=2048))
    num_periods = Column('num_periods', INTEGER(), nullable=False)

    # # relation definitions
    recurrence = relation(Recurrence,
                          primaryjoin=foreign(Recurrence.obj_guid) == guid,
                          cascade='all, delete-orphan',
                          uselist=False)

    amounts = relation('BudgetAmount',
                       back_populates="budget",
                       cascade='all, delete-orphan',
                       collection_class=CallableList,
    )


    def __str__(self):
        return "Budget<{}({}) for {} periods following pattern '{}' >".format(self.name, self.description,
                                                                               self.num_periods, self.recurrence)


class BudgetAmount(DeclarativeBase):
    """
    A GnuCash BudgetAmount

    Attributes:
        amount (:class:`decimal.Decimal`): the budgeted amount
        account (:class:`piecash.core.account.Account`): the budgeted account
        budget (:class:`Budget`): the budget of the amount
    """
    __tablename__ = 'budget_amounts'

    __table_args__ = {'sqlite_autoincrement': True}

    # column definitions
    id = Column('id', INTEGER(), primary_key=True, autoincrement=True,nullable=False)
    budget_guid = Column('budget_guid', VARCHAR(length=32),
                         ForeignKey('budgets.guid'), nullable=False)
    account_guid = Column('account_guid', VARCHAR(length=32),
                          ForeignKey('accounts.guid'), nullable=False)
    period_num = Column('period_num', INTEGER(), nullable=False)
    _amount_num = Column('amount_num', BIGINT(), nullable=False)
    _amount_denom = Column('amount_denom', BIGINT(), nullable=False)
    amount = hybrid_property_gncnumeric(_amount_num, _amount_denom)


    # relation definitions
    account = relation('Account', back_populates='budget_amounts')
    budget = relation('Budget', back_populates="amounts")

    def __str__(self):
        return "BudgetAmount<{}={}>".format(self.period_num, self.amount)





class Foo:
    def __init__(self):
        self._foo=5

    @property
    def foo(self):
        return self._foo

    @foo.setter
    def foo(self, value):
        print("setting")
        self._foo = value

f=Foo()
print(getattr(f, 'foo'))
setattr(f, 'foo', 50)
print(getattr(f, 'foo'))



# -*- coding: utf-8 -*-
import collections
import datetime

import pytest
import pytz
from sqlalchemy import create_engine, Column, TEXT
from sqlalchemy.orm import sessionmaker, composite

import piecash._common as mc
from piecash._declbase import DeclarativeBaseGuid
from piecash.business.person import Address
from piecash.sa_extra import _Date, _DateTime


def session():
    engine = create_engine("sqlite://")

    metadata = mc.DeclarativeBase.metadata
    metadata.bind = engine
    metadata.create_all()

    s = sessionmaker(bind=engine)()

    return s


class TestModelCommon(object):
    # @parametrize('helparg', ['-h', '--help'])
    def test_guid_on_declarativebase(self):
        class A(DeclarativeBaseGuid):
            __tablename__ = "a_table"

            def __init__(self):
                pass

        s = session()
        a = A()
        s.add(a)
        assert a.guid is None
        s.flush()
        assert a.guid

    def test_addr_composite(self):
        flds = "name addr1 addr2 addr3 addr4 email fax phone".split()

        class B(DeclarativeBaseGuid):
            __tablename__ = "b_table"

            def __init__(self, **kwargs):
                for k, v in kwargs.items():
                    setattr(self, k, v)

        l = []
        for fld in flds:
            col = Column(fld, TEXT())
            setattr(B, fld, col)
            l.append(col)
        B.addr = composite(Address, *l)

        s = session()
        a = B(addr1="foo")
        assert a.addr
        a.addr.fax = "baz"
        assert a.addr1 == "foo"
        assert a.addr.addr1 == "foo"
        assert a.addr.fax == "baz"
        s.add(a)
        s.flush()

    def test_date(self):
        class C(DeclarativeBaseGuid):
            __tablename__ = "c_table"
            day = Column(_Date)

            def __init__(self, day):
                self.day = day

        s = session()
        a = C(day=datetime.date(2010, 4, 12))
        s.add(a)
        s.flush()
        assert a.day

        assert str(list(s.bind.execute("select day from c_table"))[0][0]) == "20100412"

    def test_datetime(self):
        class C(DeclarativeBaseGuid):
            __tablename__ = "d_table"
            time = Column(_DateTime)

            def __init__(self, time):
                self.time = time

        s = session()
        a = C(time=datetime.datetime(2010, 4, 12, 3, 4, 5, tzinfo=pytz.utc))
        s.add(a)
        s.flush()
        assert a.time

        assert str(list(s.bind.execute("select time from d_table"))[0][0]) == "2010-04-12 03:04:05"

    def test_float_in_gncnumeric(self):
        Mock = collections.namedtuple('Mock', 'name')
        sqlcolumn_mock = Mock('')
        numeric = mc.hybrid_property_gncnumeric(sqlcolumn_mock, sqlcolumn_mock)
        with pytest.raises(TypeError) as excinfo:
            numeric.fset(None, 4020.19)
        assert ("Received a floating-point number 4020.19 where a decimal is expected. " +
                "Use a Decimal, str, or int instead") == str(excinfo.value)

    def test_weird_type_in_gncnumeric(self):
        Mock = collections.namedtuple('Mock', 'name')
        sqlcolumn_mock = Mock('')
        numeric = mc.hybrid_property_gncnumeric(sqlcolumn_mock, sqlcolumn_mock)
        with pytest.raises(TypeError) as excinfo:
            numeric.fset(None, dict())
        assert ("Received an unknown type dict where a decimal is expected. " +
                "Use a Decimal, str, or int instead") == str(excinfo.value)



# -*- coding: utf-8 -*-
from __future__ import print_function

import imp
import os
import subprocess
import sys

# # Python 2.6 subprocess.check_output compatibility. Thanks Greg Hewgill!

if "check_output" not in dir(subprocess):

    def check_output(cmd_args, *args, **kwargs):
        proc = subprocess.Popen(cmd_args, *args, stdout=subprocess.PIPE, stderr=subprocess.PIPE, **kwargs)
        out, err = proc.communicate()
        if proc.returncode != 0:
            raise subprocess.CalledProcessError(args)
        return out

    subprocess.check_output = check_output

from setuptools import setup, find_packages
from setuptools.command.test import test as TestCommand


# from setuptools.command.test import test


class PyTest(TestCommand):
    user_options = [("pytest-args=", "a", "Arguments to pass to py.test")]

    def initialize_options(self):
        TestCommand.initialize_options(self)
        self.pytest_args = []

    def finalize_options(self):
        TestCommand.finalize_options(self)
        # self.test_args = []
        # self.test_suite = True

    def run_tests(self):
        import pytest

        errno = pytest.main(self.pytest_args)
        sys.exit(errno)


try:
    import colorama

    colorama.init()  # Initialize colorama on Windows
except ImportError:
    # Don't require colorama just for running paver tasks. This allows us to
    # run `paver install' without requiring the user to first have colorama
    # installed.
    pass

# Add the current directory to the module search path.
sys.path.append(".")

# # Constants
CODE_DIRECTORY = "piecash"
DOCS_DIRECTORY = "docs"
TESTS_DIRECTORY = "tests"
DATA_DIRECTORY = "gnucash_books"
PYTEST_FLAGS = ["--doctest-modules"]

# Import metadata. Normally this would just be:
#
# from piecash import metadata
#
# However, when we do this, we also import `piecash/__init__.py'. If this
# imports names from some other modules and these modules have third-party
# dependencies that need installing (which happens after this file is run), the
# script will crash. What we do instead is to load the metadata module by path
# instead, effectively side-stepping the dependency problem. Please make sure
# metadata has no dependencies, otherwise they will need to be added to
# the setup_requires keyword.
metadata = imp.load_source("metadata", os.path.join(CODE_DIRECTORY, "metadata.py"))


# # Miscellaneous helper functions


def get_project_files():
    """Retrieve a list of project files, ignoring hidden files.

    :return: sorted list of project files
    :rtype: :class:`list`
    """
    if is_git_project():
        return get_git_project_files()

    project_files = []
    for top, subdirs, files in os.walk("."):
        for subdir in subdirs:
            if subdir.startswith("."):
                subdirs.remove(subdir)

        for f in files:
            if f.startswith("."):
                continue
            project_files.append(os.path.join(top, f))

    return project_files


def is_git_project():
    return os.path.isdir(".git")


def get_git_project_files():
    """Retrieve a list of all non-ignored files, including untracked files,
    excluding deleted files.

    :return: sorted list of git project files
    :rtype: :class:`list`
    """
    cached_and_untracked_files = git_ls_files(
        "--cached",  # All files cached in the index
        "--others",  # Untracked files
        # Exclude untracked files that would be excluded by .gitignore, etc.
        "--exclude-standard",
    )
    uncommitted_deleted_files = git_ls_files("--deleted")

    # Since sorting of files in a set is arbitrary, return a sorted list to
    # provide a well-defined order to tools like flake8, etc.
    return sorted(cached_and_untracked_files - uncommitted_deleted_files)


def git_ls_files(*cmd_args):
    """Run ``git ls-files`` in the top-level project directory. Arguments go
    directly to execution call.

    :return: set of file names
    :rtype: :class:`set`
    """
    cmd = ["git", "ls-files"]
    cmd.extend(cmd_args)
    return set(subprocess.check_output(cmd).splitlines())


def print_success_message(message):
    """Print a message indicating success in green color to STDOUT.

    :param message: the message to print
    :type message: :class:`str`
    """
    try:
        import colorama

        print(colorama.Fore.GREEN + message + colorama.Fore.RESET)
    except ImportError:
        print(message)


def print_failure_message(message):
    """Print a message indicating failure in red color to STDERR.

    :param message: the message to print
    :type message: :class:`str`
    """
    try:
        import colorama

        print(colorama.Fore.RED + message + colorama.Fore.RESET, file=sys.stderr)
    except ImportError:
        print(message, file=sys.stderr)


def read(filename):
    """Return the contents of a file.

    :param filename: file path
    :type filename: :class:`str`
    :return: the file's content
    :rtype: :class:`str`
    """
    with open(os.path.join(os.path.dirname(__file__), filename)) as f:
        return f.read()


def _lint():
    """Run lint and return an exit code."""
    # Flake8 doesn't have an easy way to run checks using a Python function, so
    # just fork off another process to do it.

    # Python 3 compat:
    # - The result of subprocess call outputs are byte strings, meaning we need
    # to pass a byte string to endswith.
    project_python_files = [filename for filename in get_project_files() if filename.endswith(b".py")]
    retcode = subprocess.call(
        ["flake8", "--ignore=E126,E121", "--max-line-length=99", "--max-complexity=10"] + project_python_files
    )
    if retcode == 0:
        print_success_message("No style errors")
    return retcode


## package dependencies
install_requires = ["SQLAlchemy>=1.0, <1.4", "SQLAlchemy-Utils!=0.36.8", "pytz", "tzlocal", "click"]
extras_require = {
    "postgres": ["psycopg2"],
    "mysql": ["PyMySQL"],
    "ledger": ["money", "babel"],
    "pandas": ["pandas"],
    "qif": ["qifparse"],
    "yahoo": ["requests"],
    "test": ["pytest", "pytest-cov", "tox"],
    "doc": ["sphinx", "sphinxcontrib-napoleon", "sphinxcontrib-programoutput", "sphinx-rtd-theme", "ipython"],
}
# build an 'all' option covering all options
extras_require["all"] = deps_all = sum(
    (extras_require[k] for k in ["postgres", "mysql", "pandas", "yahoo", "ledger"]), []
)
# add 'all' for both doc and test
extras_require["test"].extend(deps_all)
extras_require["doc"].extend(deps_all)

setup_dict = dict(
    name=metadata.package,
    version=metadata.version,
    author=metadata.authors[0],
    author_email=metadata.emails[0],
    maintainer=metadata.authors[0],
    maintainer_email=metadata.emails[0],
    url=metadata.url,
    description=metadata.description,
    long_description=read("README.rst"),
    keywords=["GnuCash", "python", "binding", "interface", "sqlalchemy"],
    license="MIT",
    platforms="any",
    # Find a list of classifiers here:
    # <http://pypi.python.org/pypi?%3Aaction=list_classifiers>
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Environment :: Console",
        "Intended Audience :: Developers",
        "Intended Audience :: Financial and Insurance Industry",
        "License :: OSI Approved :: MIT License",
        "Natural Language :: English",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Topic :: Office/Business",
        "Topic :: Office/Business :: Financial",
        "Topic :: Office/Business :: Financial :: Accounting",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ],
    packages=find_packages(exclude=(TESTS_DIRECTORY, DATA_DIRECTORY)),
    install_requires=install_requires,
    extras_require=extras_require,
    # Allow tests to be run with `python setup.py test'.
    tests_require=["pytest"] + deps_all,
    entry_points={"console_scripts": ["piecash = piecash.scripts.export:cli"]},
    cmdclass={"test": PyTest},
    test_suite="tests",
    zip_safe=False,  # don't use eggs
)


def main():
    setup(**setup_dict)


if __name__ == "__main__":
    main()



import uuid

from sqlalchemy import Column, INTEGER, BIGINT, VARCHAR, ForeignKey
from sqlalchemy.orm import composite, relation

from .person import PersonType

# change of the __doc__ string as getting error in sphinx ==> should be reported to SA project

composite.__doc__ = None  # composite.__doc__.replace(":ref:`mapper_composite`", "")

from ..sa_extra import _DateTime
from .._common import CallableList, hybrid_property_gncnumeric
from .._declbase import DeclarativeBaseGuid


class Billterm(DeclarativeBaseGuid):
    __tablename__ = 'billterms'

    __table_args__ = {}

    # column definitions
    guid = Column('guid', VARCHAR(length=32), primary_key=True, nullable=False, default=lambda: uuid.uuid4().hex)
    name = Column('name', VARCHAR(length=2048), nullable=False)
    description = Column('description', VARCHAR(length=2048), nullable=False)
    refcount = Column('refcount', INTEGER(), nullable=False)
    invisible = Column('invisible', INTEGER(), nullable=False)
    parent_guid = Column('parent', VARCHAR(length=32), ForeignKey('billterms.guid'))
    type = Column('type', VARCHAR(length=2048), nullable=False)
    duedays = Column('duedays', INTEGER())
    discountdays = Column('discountdays', INTEGER())
    _discount_num = Column('discount_num', BIGINT())
    _discount_denom = Column('discount_denom', BIGINT())
    discount = hybrid_property_gncnumeric(_discount_num, _discount_denom)
    cutoff = Column('cutoff', INTEGER())

    # relation definitions
    children = relation('Billterm',
                        back_populates='parent',
                        cascade='all, delete-orphan',
                        collection_class=CallableList,
                        )
    parent = relation('Billterm',
                      back_populates='children',
                      remote_side=guid,
                      )


class Entry(DeclarativeBaseGuid):
    __tablename__ = 'entries'

    __table_args__ = {}

    # column definitions
    date = Column('date', _DateTime(), nullable=False)
    date_entered = Column('date_entered', _DateTime())
    description = Column('description', VARCHAR(length=2048))
    action = Column('action', VARCHAR(length=2048))
    notes = Column('notes', VARCHAR(length=2048))
    _quantity_num = Column('quantity_num', BIGINT())
    _quantity_denom = Column('quantity_denom', BIGINT())
    _quantity_denom_basis = None
    quantity = hybrid_property_gncnumeric(_quantity_num, _quantity_denom)

    i_acct = Column('i_acct', VARCHAR(length=32))
    _i_price_num = Column('i_price_num', BIGINT())
    _i_price_denom = Column('i_price_denom', BIGINT())
    _i_price_denom_basis = None
    i_price = hybrid_property_gncnumeric(_i_price_num, _i_price_denom)

    _i_discount_num = Column('i_discount_num', BIGINT())
    _i_discount_denom = Column('i_discount_denom', BIGINT())
    _i_discount_denom_basis = None
    i_discount = hybrid_property_gncnumeric(_i_discount_num, _i_discount_denom)

    invoice_guid = Column('invoice', VARCHAR(length=32), ForeignKey("invoices.guid"))
    i_disc_type = Column('i_disc_type', VARCHAR(length=2048))
    i_disc_how = Column('i_disc_how', VARCHAR(length=2048))
    i_taxable = Column('i_taxable', INTEGER())
    i_taxincluded = Column('i_taxincluded', INTEGER())
    i_taxtable = Column('i_taxtable', VARCHAR(length=32))

    b_acct = Column('b_acct', VARCHAR(length=32))
    _b_price_num = Column('b_price_num', BIGINT())
    _b_price_denom = Column('b_price_denom', BIGINT())
    _b_price_denom_basis = None
    b_price = hybrid_property_gncnumeric(_b_price_num, _b_price_denom)
    bill = Column('bill', VARCHAR(length=32))
    b_taxable = Column('b_taxable', INTEGER())
    b_taxincluded = Column('b_taxincluded', INTEGER())
    b_taxtable = Column('b_taxtable', VARCHAR(length=32))
    b_paytype = Column('b_paytype', INTEGER())
    billable = Column('billable', INTEGER())
    billto_type = Column('billto_type', INTEGER())
    billto_guid = Column('billto_guid', VARCHAR(length=32))
    order_guid = Column('order_guid', VARCHAR(length=32), ForeignKey("orders.guid"))

    # relation definitions
    order = relation('Order', back_populates="entries")
    invoice = relation('Invoice', back_populates="entries")

    def __str__(self):
        return "Entry<{}>".format(self.description)


class Invoice(DeclarativeBaseGuid):
    __tablename__ = 'invoices'

    __table_args__ = {}

    # column definitions
    id = Column('id', VARCHAR(length=2048), nullable=False)
    date_opened = Column('date_opened', _DateTime())
    date_posted = Column('date_posted', _DateTime())
    notes = Column('notes', VARCHAR(length=2048), nullable=False)
    active = Column('active', INTEGER(), nullable=False)
    currency_guid = Column('currency', VARCHAR(length=32), ForeignKey('commodities.guid'), nullable=False)
    owner_type = Column('owner_type', INTEGER())
    owner_guid = Column('owner_guid', VARCHAR(length=32))

    # owner = generic_relationship(owner_type, owner_guid,
    #                              map_type2discriminator={"Customer": 2,
    #                                                      "Vendor": 1,
    #                                                      })
    term_guid = Column('terms', VARCHAR(length=32), ForeignKey('billterms.guid'))
    billing_id = Column('billing_id', VARCHAR(length=2048))
    post_txn_guid = Column('post_txn', VARCHAR(length=32), ForeignKey('lots.guid'))
    post_lot_guid = Column('post_lot', VARCHAR(length=32), ForeignKey('transactions.guid'))
    post_acc_guid = Column('post_acc', VARCHAR(length=32), ForeignKey('accounts.guid'))
    billto_type = Column('billto_type', INTEGER())
    billto_guid = Column('billto_guid', VARCHAR(length=32))
    _charge_amt_num = Column('charge_amt_num', BIGINT())
    _charge_amt_denom = Column('charge_amt_denom', BIGINT())
    charge_amt = hybrid_property_gncnumeric(_charge_amt_num, _charge_amt_denom)

    # relation definitions
    # todo: check all relations and understanding of types...
    term = relation('Billterm')
    currency = relation('Commodity')
    post_account = relation('Account')
    post_lot = relation('Lot')
    post_txn = relation('Transaction')

    entries = relation('Entry',
                       back_populates='invoice',
                       cascade='all, delete-orphan',
                       collection_class=CallableList,
                       )

    def __str__(self):
        return "Invoice<{}>".format(self.id)


class Job(DeclarativeBaseGuid):
    __tablename__ = 'jobs'

    __table_args__ = {}

    # column definitions
    id = Column('id', VARCHAR(length=2048), nullable=False)
    name = Column('name', VARCHAR(length=2048), nullable=False)
    reference = Column('reference', VARCHAR(length=2048), nullable=False)
    active = Column('active', INTEGER(), nullable=False)
    owner_type = Column('owner_type', INTEGER())
    owner_guid = Column('owner_guid', VARCHAR(length=32))

    # relation definitions
    # todo: owner_guid/type links to Vendor or Customer
    def __init__(self, name, owner, reference="", active=1):
        self.name = name
        self.reference = reference
        self.active = active
        self.owner_type = PersonType[type(owner)]
        self.owner_guid = owner.guid
        if owner.book:
            owner.book.add(self)

    def __str__(self):
        return "Job<{self.name}>".format(self=self)

    def on_book_add(self):
        self._assign_id()

    # hold the name of the counter to use for id
    _counter_name = "counter_job"

    def _assign_id(self):
        if not self.id:
            cnt = getattr(self.book, self._counter_name) + 1
            setattr(self.book, self._counter_name, cnt)
            self.id = "{:06d}".format(cnt)


# This class exists in code but not in the GUI (to confirm?)

class Order(DeclarativeBaseGuid):
    __tablename__ = 'orders'

    __table_args__ = {}

    # column definitions
    id = Column('id', VARCHAR(length=2048), nullable=False)
    notes = Column('notes', VARCHAR(length=2048), nullable=False)
    reference = Column('reference', VARCHAR(length=2048), nullable=False)
    active = Column('active', INTEGER(), nullable=False)

    date_opened = Column('date_opened', _DateTime(), nullable=False)
    date_closed = Column('date_closed', _DateTime(), nullable=False)
    owner_type = Column('owner_type', INTEGER(), nullable=False)
    owner_guid = Column('owner_guid', VARCHAR(length=32), nullable=False)

    # relation definitions
    # todo: owner_guid/type links to Vendor or Customer
    entries = relation('Entry',
                       back_populates='order',
                       cascade='all, delete-orphan',
                       collection_class=CallableList,
                       )



import sqlite3, os
import click


def dump(name):
    con = sqlite3.connect(name)
    with open(f'{name}.sql', 'w') as f:
        for line in con.iterdump():
            f.write('%s\n' % line.replace("VARCHAR(","TEXT("))

import piecash

b = piecash.create_book("default_piecash.gnucash",overwrite=True)
b.save()

#dump("default_2_21.gnucash")
dump("default_piecash.gnucash")



from __future__ import unicode_literals

import re
from functools import singledispatch
from locale import getdefaultlocale

from .core import Transaction, Account, Commodity, Price, Book

"""original script from https://github.com/MatzeB/pygnucash/blob/master/gnucash2ledger.py by Matthias Braun matze@braunis.de
 adapted for:
 - python 3 support
 - new string formatting
"""

try:
    from money import Money
except ImportError:
    Money = None


@singledispatch
def ledger(obj, **kwargs):
    raise NotImplemented


CURRENCY_RE = re.compile("^[A-Z]{3}$")
NUMBER_RE = re.compile("[0-9\., ]")


def format_commodity(mnemonic, locale):
    if CURRENCY_RE.match(mnemonic):
        # format the commodity
        s = format_currency(0, 0, mnemonic, locale)

        # remove the non currency part and real white spaces
        return NUMBER_RE.sub("", s)
    else:
        if NUMBER_RE.search(mnemonic):
            return '"{}"'.format(mnemonic)
        else:
            return mnemonic

    return NUMBER_RE.sub("", s)


def format_currency(amount, decimals, currency, locale=False):
    if locale:
        if locale is True:
            locale = getdefaultlocale()[0]
        if Money is None:
            raise ValueError(
                f"You must install Money ('pip install money') to export to ledger in your locale '{locale}'"
            )
        return Money(amount=amount, currency=currency).format(locale)
    else:
        if Money:
            try:
                # version from Money
                return str(Money(amount=amount, currency=currency))
            except ValueError:
                # local hand made version
                return "{:.{}f} {}".format(amount, decimals, format_commodity(currency, locale))
        else:
            # local hand made version
            return "{:.{}f} {}".format(amount, decimals, currency)


@ledger.register(Transaction)
def _(tr, locale=False, **kwargs):
    """Return a ledger-cli alike representation of the transaction"""
    s = [
        "{:%Y-%m-%d} {}{}\n".format(
            tr.post_date, "({}) ".format(tr.num.replace(")", "")) if tr.num else "", tr.description
        )
    ]
    if tr.notes:
        s.append("\t;{}\n".format(tr.notes))
    for split in sorted(
        tr.splits, key=lambda split: (split.value, split.transaction_guid, split.account_guid)
    ):
        if split.account.commodity.mnemonic == "template":
            return ""
        if split.reconcile_state in ["c", "y"]:
            s.append("\t* {:38}  ".format(split.account.fullname))
        else:
            s.append("\t{:40}  ".format(split.account.fullname))
        if split.account.commodity != tr.currency:
            s.append(
                "{quantity} @@ {amount}".format(
                    quantity=format_currency(
                        split.quantity,
                        split.account.commodity.precision,
                        split.account.commodity.mnemonic,
                        locale=False,
                    ),
                    amount=format_currency(
                        abs(split.value), tr.currency.precision, tr.currency.mnemonic, locale
                    ),
                )
            )
        else:
            s.append(
                format_currency(split.value, tr.currency.precision, tr.currency.mnemonic, locale)
            )

        if split.memo:
            s.append(" ;   {:20}".format(split.memo))
        s.append("\n")

    return "".join(s)


@ledger.register(Commodity)
def _(cdty, locale=False, commodity_notes=False, **kwargs):
    """Return a ledger-cli alike representation of the commodity"""
    if cdty.mnemonic in ["", "template"]:
        return ""
    res = "commodity {}\n".format(format_commodity(cdty.mnemonic, locale))
    if cdty.fullname != "" and commodity_notes:
        res += "\tnote {}\n".format(cdty.fullname, locale)
    res += "\n"
    return res


@ledger.register(Account)
def _(acc, short_account_names=False, **kwargs):
    """Return a ledger-cli alike representation of the account"""
    # ignore "dummy" accounts
    if acc.type is None or acc.parent is None:
        return ""
    if acc.commodity.mnemonic == "template":
        return ""

    if short_account_names:
        res = "account {}\n".format(acc.name)
    else:
        res = "account {}\n".format(acc.fullname)

    if acc.description != "":
        res += "\tnote {}\n".format(acc.description)

    res += '\tcheck commodity == "{}"\n'.format(acc.commodity.mnemonic)  # .replace('"', '\\"'))
    return res


@ledger.register(Price)
def _(price, locale=False, **kwargs):
    """Return a ledger-cli alike representation of the price"""
    return "P {:%Y-%m-%d %H:%M:%S} {} {}\n".format(
        price.date,
        format_commodity(price.commodity.mnemonic, locale),
        format_currency(price.value, price.currency.precision, price.currency.mnemonic, locale),
    )


@ledger.register(Book)
def _(book, **kwargs):
    """Return a ledger-cli alike representation of the book"""
    res = []

    # Commodities
    for commodity in sorted(book.commodities, key=lambda cdty: cdty.mnemonic):
        res.append(ledger(commodity, **kwargs))

    # Accounts
    if kwargs.get("short_account_names"):  # check that no ambiguity in account names
        accounts = [acc.name for acc in book.accounts]
        if len(accounts) != len(set(accounts)):
            raise ValueError(
                "You have duplicate short names in your book. "
                "You cannot use the 'short_account_names' option."
            )
    for acc in book.accounts:
        res.append(ledger(acc, **kwargs))
        res.append("\n")

    # Prices
    for price in sorted(book.prices, key=lambda x: (x.commodity_guid, x.currency_guid, x.date)):
        res.append(ledger(price, **kwargs))
    res.append("\n")

    for trans in sorted(book.transactions, key=lambda x: (x.currency_guid, x.post_date)):
        res.append(ledger(trans, **kwargs))
        res.append("\n")

    return "".join(res)



import warnings
from collections import defaultdict
from operator import attrgetter

from sqlalchemy import Column, VARCHAR, ForeignKey
from sqlalchemy.orm import relation, aliased, joinedload
from sqlalchemy.orm.base import instance_state
from sqlalchemy.orm.exc import NoResultFound

from . import factories
from .account import Account
from .commodity import Commodity, Price
from .transaction import Split, Transaction
from .._common import CallableList, GnucashException
from .._declbase import DeclarativeBaseGuid
from ..business.invoice import Invoice
from ..sa_extra import kvp_attribute


class Book(DeclarativeBaseGuid):
    """
    A Book represents a GnuCash document. It is created through one of the two factory functions
    :func:`create_book` and :func:`open_book`.

    Canonical use is as a context manager like (the book is automatically closed at the end of the with block)::

        with create_book() as book:
            ...

    .. note:: If you do not use the context manager, do not forget to close the session explicitly (``book.close()``)
       to release any lock on the file/DB.

    The book puts at disposal several attributes to access the main objects of the GnuCash document::

        # to get the book and the root_account
        ra = book.root_account

        # to get the list of accounts, commodities or transactions
        for acc in book.accounts:  # or book.commodities or book.transactions
            # do something with acc

        # to get a specific element of these lists
        EUR = book.commodities(namespace="CURRENCY", mnemonic="EUR")

        # to get a list of all objects of some class (even non core classes)
        budgets = book.get(Budget)
        # or a specific object
        budget = book.get(Budget, name="my first budget")

    You can check a session has changes (new, deleted, changed objects) by getting the ``book.is_saved`` property.
    To save or cancel changes, use ``book.save()`` or ``book.cancel()``::

        # save a session if it is no saved (saving a unchanged session is a no-op)
        if not book.is_saved:
            book.save()

    Attributes:
        root_account (:class:`piecash.core.account.Account`): the root account of the book
        root_template (:class:`piecash.core.account.Account`): the root template of the book (usage not yet clear...)
        default_currency (:class:`piecash.core.commodity.Commodity`): the currency of the root account (=default currency of the book)
        uri (str): connection string of the book (set by the GncSession when accessing the book)
        session (:class:`sqlalchemy.orm.session.Session`): the sqlalchemy session encapsulating the book
        use_trading_accounts (bool): true if option "Use trading accounts" is enabled
        use_split_action_field (bool): true if option "Use Split Action Field for Number" is enabled
        RO_threshold_day (int): value of Day Threshold for Read-Only Transactions (red line)
        control_mode (list(str)) : list of allowed non-standard operations like : "allow-root-subaccounts"
        counter_customer (int) : counter for :class:`piecash.business.person.Customer` id (link to slot "counters/gncCustomer")
        counter_vendor (int) : counter for :class:`piecash.business.person.Vendor` id (link to slot "counters/gncVendor")
        counter_employee (int) : counter for :class:`piecash.business.person.Employee` id (link to slot "counters/gncEmployee")
        counter_invoice (int) : counter for :class:`piecash.business.invoice.Invoice` id (link to slot "counters/gncInvoice")
        counter_job (int) : counter for :class:`piecash.business.invoice.Job` id (link to slot "counters/gncJob")
        counter_bill (int) : counter for :class:`piecash.business.invoice.Bill` id (link to slot "counters/gncBill")
        counter_exp_voucher (int) : counter for :class:`piecash.business.invoice.Invoice` id (link to slot "counters/gncExpVoucher")
        counter_order (int) : counter for :class:`piecash.business.invoice.Order` id (link to slot "counters/gncOrder")
        business_company_phone (str): phone number of book company (link to slit "options/Business/Company Phone Number")
        business_company_email (str): email of book company (link to slit "options/Business/Company Email Address")
        business_company_contact (str): contact person of book company (link to slit "options/Business/Company Contact Person")
        business_company_ID (str): ID of book company (link to slit "options/Business/Company ID")
        business_company_name (str): name of book company (link to slit "options/Business/Company Name")
        business_company_address (str): address of book company (link to slit "options/Business/Company Address")
        business_company_website (str): website URL of book company (link to slit "options/Business/Company Website URL")
    """
    __tablename__ = 'books'

    __table_args__ = {}

    # column definitions
    root_account_guid = Column('root_account_guid', VARCHAR(length=32),
                               ForeignKey('accounts.guid'), nullable=False)
    root_template_guid = Column('root_template_guid', VARCHAR(length=32),
                                ForeignKey('accounts.guid'), nullable=False)

    # relation definitions
    root_account = relation('Account',
                            # back_populates='root_book',
                            foreign_keys=[root_account_guid],
                            )
    root_template = relation('Account',
                             foreign_keys=[root_template_guid])

    uri = None
    session = None

    # link options to KVP
    use_trading_accounts = kvp_attribute("options/Accounts/Use Trading Accounts",
                                         from_gnc=lambda v: v == 't',
                                         to_gnc=lambda v: 't',
                                         default=False)

    use_split_action_field = kvp_attribute("options/Accounts/Use Split Action Field for Number",
                                           from_gnc=lambda v: v == 't',
                                           to_gnc=lambda v: 't' if v else 'f',
                                           default=False)

    RO_threshold_day = kvp_attribute("options/Accounts/Day Threshold for Read-Only Transactions (red line)",
                                     from_gnc=lambda v: int(v),
                                     to_gnc=lambda v: float(v),
                                     default=0)

    counter_customer = kvp_attribute("counters/gncCustomer", default=0)
    counter_vendor = kvp_attribute("counters/gncVendor", default=0)
    counter_employee = kvp_attribute("counters/gncEmployee", default=0)
    counter_invoice = kvp_attribute("counters/gncInvoice", default=0)
    counter_job = kvp_attribute("counters/gncJob", default=0)
    counter_bill = kvp_attribute("counters/gncBill", default=0)
    counter_exp_voucher = kvp_attribute("counters/gncExpVoucher", default=0)
    counter_order = kvp_attribute("counters/gncOrder", default=0)

    business_company_phone = kvp_attribute("options/Business/Company Phone Number", default="")
    business_company_email = kvp_attribute("options/Business/Company Email Address", default="")
    business_company_contact = kvp_attribute("options/Business/Company Contact Person", default="")
    business_company_ID = kvp_attribute("options/Business/Company ID", default="")
    business_company_name = kvp_attribute("options/Business/Company Name", default="")
    business_company_address = kvp_attribute("options/Business/Company Address", default="")
    business_company_website = kvp_attribute("options/Business/Company Website URL", default="")

    def __init__(self, root_account=None, root_template=None):
        self.root_account = root_account
        self.root_template = root_template

    def __str__(self):
        return "Book<{}>".format(self.uri)

    _control_mode = None

    @property
    def control_mode(self):
        if self._control_mode is None:
            self._control_mode = []
        return self._control_mode

    @property
    def default_currency(self):
        return self.root_account.commodity

    @default_currency.setter
    def default_currency(self, value):
        assert isinstance(value, Commodity) and value.namespace == "CURRENCY"

        self.root_account.commodity = value

    @property
    def book(self):
        warnings.warn("deprecated", DeprecationWarning)
        return self

    def validate(self):
        Book.validate_book(self.session)

    @staticmethod
    def track_dirty(session, flush_context, instances):
        """
        Record in session._all_changes the objects that have been modified before each flush
        """
        for change, l in {"dirty": session.dirty,
                          "new": session.new,
                          "deleted": session.deleted}.items():
            for obj in l:
                # retrieve the dictionnary of changes for the given obj
                attrs = session._all_changes.setdefault(id(obj), {})
                # add the change of state to the list of state changes
                attrs.setdefault("STATE_CHANGES", []).append(change)
                attrs.setdefault("OBJECT", obj)
                # save old value of attr if not already saved
                # (if a value is changed multiple time, we keep only the first "old value")
                for k, v in instance_state(obj).committed_state.items():
                    if k not in attrs:
                        attrs[k] = v

    @staticmethod
    def validate_book(session):
        session.flush()

        # identify object to validate
        txs = set()

        # iterate on all explicitly changes objects to see
        # if we need to add other objects for check
        for attrs in session._all_changes.values():
            obj = attrs["OBJECT"]
            for o_to_validate in obj.object_to_validate(attrs["STATE_CHANGES"]):
                txs.add(o_to_validate)

        assert None not in txs, "No object should return None to validate. fix the code"

        # sort object from local to global (ensure Split checked before Transaction)
        from . import Account, Transaction, Split, Commodity
        sort_order = defaultdict(lambda: 20, {Account: 10, Transaction: 5, Split: 3, Commodity: 2})
        txs = list(txs)
        txs.sort(key=lambda x: sort_order[type(x)])

        # for each object, validate it
        for tx in txs:
            tx.validate()

    _trading_accounts = None

    def trading_account(self, cdty):
        """Return the trading account related to the commodity. If it does not exist and the option
        "Use Trading Accounts" is enabled, create it on the fly"""
        key = namespace, mnemonic = cdty.namespace, cdty.mnemonic
        if self._trading_accounts is None:
            self._trading_accounts = {}

        tacc = self._trading_accounts.get(key, None)
        if tacc: return tacc

        from .account import Account

        try:
            trading = self.root_account.children(name="Trading")
        except KeyError:
            trading = Account(name="Trading",
                              type="TRADING",
                              placeholder=1,
                              commodity=self.default_currency,
                              parent=self.root_account)
        try:
            nspc = trading.children(name=cdty.namespace)
        except KeyError:
            nspc = Account(name=namespace,
                           type="TRADING",
                           placeholder=1,
                           commodity=self.default_currency,
                           parent=trading)
        try:
            tacc = nspc.children(name=cdty.mnemonic)
        except KeyError:
            tacc = Account(name=mnemonic,
                           type="TRADING",
                           placeholder=0,
                           commodity=cdty,
                           parent=nspc)
        # self.flush()
        return tacc

    # add session alike functions
    def add(self, obj):
        """Add an object to the book (to be used if object not linked in any way to the book)"""
        self.session.add(obj)
        obj.on_book_add()

    def delete(self, obj):
        """Delete an object from the book (to remove permanently an object)"""
        self.session.delete(obj)

    def save(self):
        """Save the changes to the file/DB (=commit transaction)
        """
        self.session.commit()

    def flush(self):
        """Flush the book"""
        self.session.flush()

    def cancel(self):
        """Cancel all the changes that have not been saved (=rollback transaction)
        """
        self.session.rollback()

    @property
    def is_saved(self):
        """Are all the changes saved to the file/DB?

        You can check a session has changes (new, deleted, changed objects)
        by getting the ``book.is_saved`` property.
        """
        return self.session.is_saved

    # add context manager that close the session when leaving
    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    def close(self):
        """Close a session. Any changes not yet saved are rolled back. Any lock on the file/DB is released.
        """
        session = self.session
        # cancel pending changes
        session.rollback()
        # if self._acquire_lock:
        # # remove the lock
        # session.delete_lock()
        session.close()

    # add general getters for gnucash classes
    def get(self, cls, **kwargs):
        """
        Generic getter for a GnuCash object in the `GncSession`. If no kwargs is given, it returns the list of all
        objects of type cls (uses the sqlalchemy session.query(cls).all()).
        Otherwise, it gets the unique object which attributes match the kwargs
        (uses the sqlalchemy session.query(cls).filter_by(\*\*kwargs).one() underneath)::

            # to get the first account with name="Income"
            inc_account = session.get(Account, name="Income")

            # to get all accounts
            accs = session.get(Account)

        Args:
            cls (class): the class of the object to retrieve (Account, Price, Budget,...)
            kwargs (dict): the attributes to filter on

        Returns:
            object: the unique object if it exists, raises exceptions otherwise
        """
        if kwargs:
            try:
                return self.session.query(cls).filter_by(**kwargs).one()
            except NoResultFound:
                raise ValueError("Could not find a {}({})".format(cls.__name__,
                                                                  kwargs))
        else:
            return self.session.query(cls)

    @property
    def transactions(self):
        """
        gives easy access to all transactions in the book through a :class:`piecash.model_common.CallableList`
        of :class:`piecash.core.transaction.Transaction`
        """
        from .transaction import Transaction

        return CallableList(self.session.query(Transaction))

    @property
    def splits(self):
        """
        gives easy access to all splits in the book through a :class:`piecash.model_common.CallableList`
        of :class:`piecash.core.transaction.Split`
        """
        from .transaction import Split

        return CallableList(self.session.query(Split))

    @property
    def accounts(self):
        """
        gives easy access to all accounts in the book through a :class:`piecash.model_common.CallableList`
        of :class:`piecash.core.account.Account`
        """
        from .account import Account

        return CallableList(self.session.query(Account).filter(Account.parent != None))

    @property
    def commodities(self):
        """
        gives easy access to all commodities in the book through a :class:`piecash.model_common.CallableList`
        of :class:`piecash.core.commodity.Commodity`
        """
        from .commodity import Commodity

        return CallableList(self.session.query(Commodity))

    @property
    def invoices(self):
        """
        gives easy access to all commodities in the book through a :class:`piecash.model_common.CallableList`
        of :class:`piecash.core.commodity.Commodity`
        """

        return CallableList(self.session.query(Invoice))

    @property
    def currencies(self):
        """
        gives easy access to all currencies in the book through a :class:`piecash.model_common.CallableList`
        of :class:`piecash.core.commodity.Commodity`
        """
        from .commodity import Commodity

        def fallback(mnemonic):
            cur = factories.create_currency_from_ISO(isocode=mnemonic)
            self.add(cur)
            self.flush()
            return cur

        cl = CallableList(self.session.query(Commodity).filter_by(namespace="CURRENCY"))
        cl.fallback = fallback
        return cl

    @property
    def prices(self):
        """
        gives easy access to all prices in the book through a :class:`piecash.model_common.CallableList`
        of :class:`piecash.core.commodity.Price`
        """
        from .commodity import Price

        return CallableList(self.session.query(Price))

    @property
    def customers(self):
        """
        gives easy access to all commodities in the book through a :class:`piecash.model_common.CallableList`
        of :class:`piecash.business.people.Customer`
        """
        from ..business import Customer

        return CallableList(self.session.query(Customer))

    @property
    def vendors(self):
        """
        gives easy access to all commodities in the book through a :class:`piecash.model_common.CallableList`
        of :class:`piecash.business.people.Vendor`
        """
        from ..business import Vendor

        return CallableList(self.session.query(Vendor))

    @property
    def employees(self):
        """
        gives easy access to all commodities in the book through a :class:`piecash.model_common.CallableList`
        of :class:`piecash.business.people.Employee`
        """
        from ..business import Employee

        return CallableList(self.session.query(Employee))

    @property
    def taxtables(self):
        """
        gives easy access to all commodities in the book through a :class:`piecash.model_common.CallableList`
        of :class:`piecash.business.tax.Taxtable`
        """
        from ..business import Taxtable

        return CallableList(self.session.query(Taxtable))

    @property
    def query(self):
        """
        proxy for the query function of the underlying sqlalchemy session
        """
        return self.session.query

    def preload(self):
        # preload list of accounts
        accounts = self.session.query(Account).options(joinedload("splits").joinedload("transaction"),
                                                       joinedload("children"),
                                                       joinedload("commodity"),
                                                       ).all()

        # load all splits
        splits = self.session.query(Split).join(Transaction).options(
            joinedload("account"),
            joinedload("lot")) \
            .order_by(Transaction.post_date, Split.value).all()

        return accounts, splits

    def splits_df(self, additional_fields=None):
        """
        Return a pandas DataFrame with all splits (:class:`piecash.core.commodity.Split`) from the book
        
        :parameters: :class:`list`

        :return: :class:`pandas.DataFrame`
        """
        try:
            import pandas
        except ImportError:
            raise GnucashException("pandas is required to output dataframes")

        # Initialise default argument here
        additional_fields = additional_fields if additional_fields else []

        # preload list of accounts
        accounts = self.session.query(Account).all()

        # preload list of commodities
        commodities = self.session.query(Commodity).filter(Commodity.namespace != "template").all()

        # preload list of transactions
        transactions = self.session.query(Transaction).all()

        # load all splits
        splits = self.session.query(Split).join(Transaction) \
            .order_by(Transaction.post_date, Split.value).all()

        # build dataframe. Adds additional transaction.guid field.
        fields = ["guid", "value", "quantity", "memo", "transaction.guid", "transaction.description",
                  "transaction.post_date", "transaction.currency.guid", "transaction.currency.mnemonic",
                  "account.fullname", "account.commodity.guid", "account.commodity.mnemonic",
                  ] + additional_fields
        fields_getter = [attrgetter(fld) for fld in fields]
        df_splits = pandas.DataFrame([[fg(sp) for fg in fields_getter]
                                      for sp in splits], columns=fields)
        df_splits = df_splits[df_splits["account.commodity.mnemonic"] != "template"]
        df_splits = df_splits.set_index("guid")

        return df_splits

    def prices_df(self):
        """
        Return a pandas DataFrame with all prices (:class:`piecash.core.commodity.Price`) from the book

        :return: :class:`pandas.DataFrame`
        """
        try:
            import pandas
        except ImportError:
            raise GnucashException("pandas is required to output dataframes")

        # preload list of commodities
        commodities = self.session.query(Commodity).all()

        # load all prices
        Currency = aliased(Commodity)
        prices = self.session.query(Price) \
            .join(Commodity, Price.commodity) \
            .join(Currency, Price.currency) \
            .order_by(Commodity.mnemonic, Price.date, Currency.mnemonic).all()

        fields = ["date", "type", "value",
                  "commodity.guid", "commodity.mnemonic",
                  "currency.guid", "currency.mnemonic", ]
        fields_getter = [attrgetter(fld) for fld in fields]
        df_prices = pandas.DataFrame([[fg(pr) for fg in fields_getter]
                                      for pr in prices], columns=fields)

        return df_prices



from decimal import Decimal

from piecash import ledger
from test_helper import (db_sqlite_uri, db_sqlite, new_book, new_book_USD, book_uri,
                         book_transactions, book_complex)

# dummy line to avoid removing unused symbols
a = db_sqlite_uri, db_sqlite, new_book, new_book_USD, book_uri, book_transactions, book_complex


def test_get_balance(book_complex):
    """
    Tests listing the commodity quantity in the account.
    """

    asset = book_complex.accounts.get(name="Asset")
    broker = book_complex.accounts.get(name="Broker")
    foo_stock = book_complex.accounts.get(name="Foo stock")
    expense =  book_complex.accounts.get(name="Expense")
    income = book_complex.accounts.get(name="Income")
    assert foo_stock.get_balance(recurse=True) == Decimal('130')
    assert broker.get_balance(recurse=True) == Decimal('117')
    assert asset.get_balance(recurse=False) == Decimal('0')
    assert asset.get_balance() == Decimal('24695.3')
    assert expense.get_balance() == Decimal('260')
    assert income.get_balance() == Decimal('150')
    assert income.get_balance(natural_sign=False) == Decimal('-150')



# coding=utf-8
from __future__ import unicode_literals

# dummy line to avoid removing unused symbols
from piecash import Employee, Job
from test_helper import db_sqlite_uri, db_sqlite, new_book, new_book_USD, book_uri, book_basic, Person

a = db_sqlite_uri, db_sqlite, new_book, new_book_USD, book_uri, book_basic, Person


class TestBusinessPerson_create_Person(object):
    """
    Person is a parameter taking values in [Customer, Vendor, Employee]
    """

    def test_create_customer_job(self, book_basic, Person):
        EUR = book_basic.commodities(namespace="CURRENCY")

        # create detached person
        c = Person(name="John Föo", currency=EUR)
        book_basic.add(c)

        if Person != Employee:
            j = Job(name="my job", owner=c)

        book_basic.validate()
        book_basic.flush()
        if Person != Employee:
            print(c.jobs)
